# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
import time

from ...constants import MAX_PAGE_SIZE, PAGE_SIZE
from ...exceptions import JsonError, NotFoundError
from ...tools import dt_now, dt_sec_ago, listify
from ..adapters import Adapters
from ..asset_callbacks import get_callbacks_cls
from ..mixins import ModelMixins, PagingMixins
from .fields import Fields
from .labels import Labels
from .saved_query import SavedQuery


class AssetMixin(ModelMixins, PagingMixins):
    """API model for working with user and device assets."""

    FIELD_TAGS = "labels"
    FIELD_AXON_ID = "internal_axon_id"
    FIELD_ADAPTERS = "adapters"
    FIELD_ADAPTER_LEN = "adapter_list_length"

    FIELDS_API = [FIELD_AXON_ID, FIELD_ADAPTERS, FIELD_TAGS, FIELD_ADAPTER_LEN]

    @property
    def fields_default(self):
        """Fields to add to all get calls for this asset type.

        Returns:
            :obj:`list` of :obj:`dict`: fields to add to
        """
        raise NotImplementedError  # pragma: no cover

    def destroy(self, destroy, history):
        """Destroy ALL assets."""
        return self._destroy(destroy=destroy, history=history)

    def count(self, query=None):
        """Get the count of assets.

        Args:
            query (:obj:`str`, optional): default ``None`` -

                * if ``None`` return the count of all assets
                * if :obj:`str` return the count of assets that match a
                  query built by the GUI query wizard

        Returns:
            :obj:`int`: count of assets matching query
        """
        return self._count(query=query)

    def count_by_saved_query(self, name):
        """Get the count of assets that would be returned by a saved query.

        Args:
            name (:obj:`str`): name of saved query to get count of assets from

        Returns:
            :obj:`int`: count of assets matching query in saved query
        """
        sq = self.saved_query.get_by_name(value=name)
        return self._count(query=sq["view"]["query"]["filter"])

    def get_generator(
        self,
        query=None,
        fields=None,
        fields_manual=None,
        fields_regex=None,
        fields_default=True,
        fields_map=None,
        max_rows=None,
        max_pages=None,
        page_size=MAX_PAGE_SIZE,
        page_start=0,
        page_sleep=0,
        use_cursor=True,
        export=None,
        **kwargs,
    ):
        """Get an iterator of objects for a given query using paging.

        Args:
            query (:obj:`str`, optional): default ``None`` -

                * if ``None`` return all assets
                * if :obj:`str` return the assets that match a query built
                  by the GUI query wizard
            fields (:obj:`list` of :obj:`str`, optional): default ``None`` -
                the fields to include for each asset, will be validated and
                processed into their fully qualified name using
                :meth:`Fields.validate`
            fields_manual (:obj:`list` of :obj:`str`, optional): default ``None`` -
                list of fully qualified fields to include for each asset
            fields_regex (:obj:`list` of :obj:`str`, optional): default ``None`` -
                list of fields to add using regular expression matches, will be
                validated and process into the matching fully qualified names using
                :meth:`Fields.validate`
            fields_default (:obj:`bool`, optional): default ``True`` -
                Include the fields in _default_fields
            fields_error (:obj:`bool`, optional): default ``True`` -
                throw an exception if fields fail to be validated by
                :meth:`Fields.validate`
            max_rows (:obj:`int`, optional): default ``None`` - return N assets
            max_pages (:obj:`int`, optional): default ``None`` - return N pages of assets
            page_size (:obj:`int`, optional):
                default default :data:`MAX_PAGE_SIZE` -
                return N assets per page
            page_start (:obj:`int`, optional): default ``0`` - start at page N

        Yields:
            :obj:`dict`: asset matching **query**
        """
        if getattr(self, "NO_CURSOR_OVERRIDE", False):
            use_cursor = False

        fields_map = fields_map or self.fields.get()

        fields = self.fields.validate(
            fields=fields,
            fields_manual=fields_manual,
            fields_regex=fields_regex,
            fields_default=fields_default,
            fields_map=fields_map,
        )

        page_size = self._get_page_size(page_size=page_size, max_rows=max_rows)

        store = {
            "query": query,
            "fields": fields,
        }

        state = {
            "max_pages": max_pages,
            "max_rows": max_rows,
            "page": {},
            "page_cursor": None,
            "page_left": 0,
            "page_num": page_start,
            "page_size": page_size,
            "page_sleep": page_sleep,
            "page_total": 0,
            "rows_fetched": 0,
            "rows_page": 0,
            "rows_processed": 0,
            "rows_start": page_start * page_size,
            "rows_total": 0,
            "time_fetch": 0,
            "time_page": 0,
            "use_cursor": use_cursor,
        }

        callbacks_cls = get_callbacks_cls(export=export)

        self._LAST_CALLBACKS = callbacks = callbacks_cls(
            apiobj=self,
            fields_map=fields_map,
            getargs=kwargs,
            fields=fields,
            query=query,
            state=state,
        )

        callbacks.start()

        self.LOG.info(f"START FETCH store={store}")

        while True:
            rows = self._get_page(state=state, store=store)

            for row in rows:
                row_items = callbacks.row(row=row)

                state["rows_processed"] += 1

                for row_item in listify(obj=row_items):
                    yield row_item

                if self._get_page_stop_rows(
                    row=row, state=state, store=store, rows=rows
                ):
                    break

            if self._get_page_stop_fetch(rows=rows, state=state, store=store):
                callbacks.stop()
                break

            time.sleep(state["page_sleep"])

        self.LOG.info(f"FINISH FETCH store={store}")

    def get_by_id(self, id):
        """Get the full metadata of all adapters for a single asset.

        Args:
            id (:obj:`str`): internal_axon_id of asset to get all metadata for

        Raises:
            :exc:`ValueNotFoundError`: if asset is not found with supplied **id**

        Returns:
            :obj:`dict`: dict with all metadata for all adapters for asset with
                **id** of internal_axon_id
        """
        try:
            return self._get_by_id(id=id)
        except JsonError:
            otype = self.router._object_type
            msg = f"Failed to find internal_axon_id {id!r} for {otype}"
            raise NotFoundError(msg)

    def get_by_saved_query(self, name, **kwargs):
        """Get assets that would be returned by a saved query.

        Args:
            name (:obj:`str`): name of saved query to get count of assets from
            **kwargs: passed to :meth:`get`

        Yields:
            :obj:`dict`: asset matching **query** if generator is True

        Returns:
            :obj:`list` of :obj:`dict`: assets matching **query** if generator is False
        """
        sq = self.saved_query.get_by_name(value=name)
        kwargs["query"] = sq["view"]["query"]["filter"]
        kwargs["fields_manual"] = sq["view"]["fields"]
        kwargs.setdefault("fields_default", False)
        return self.get(**kwargs)

    def get_by_values(
        self,
        values,
        field,
        not_flag=False,
        pre="",
        post="",
        field_manual=False,
        fields_map=None,
        **kwargs,
    ):
        """Pass."""
        if not field_manual:
            fields_map = fields_map or self.fields.get()

        field = self.fields.get_field_name(
            value=field, fields_map=fields_map, field_manual=field_manual
        )

        match = listify(values)
        match = [f"'{x.strip()}'" for x in match]
        match = ", ".join(match)

        inner = f"{field} in [{match}]"

        kwargs["query"] = self._build_query(
            inner=inner, pre=pre, post=post, not_flag=not_flag,
        )

        return self.get(fields_map=fields_map, **kwargs)

    def get_by_value_regex(
        self,
        value,
        field,
        cast_insensitive=True,
        not_flag=False,
        pre="",
        post="",
        field_manual=False,
        fields_map=False,
        **kwargs,
    ):
        """Pass."""
        if not field_manual:
            fields_map = fields_map or self.fields.get()

        field = self.fields.get_field_name(
            value=field, fields_map=fields_map, field_manual=field_manual
        )

        if cast_insensitive:
            inner = f'{field} == regex("{value}", "i")'
        else:
            inner = f'{field} == regex("{value}")'

        kwargs["query"] = self._build_query(
            inner=inner, pre=pre, post=post, not_flag=not_flag,
        )

        return self.get(fields_map=fields_map, **kwargs)

    def get_by_value(
        self,
        value,
        field,
        not_flag=False,
        pre="",
        post="",
        field_manual=False,
        fields_map=False,
        **kwargs,
    ):
        """Build query to get an asset by field value."""
        if not field_manual:
            fields_map = fields_map or self.fields.get()

        field = self.fields.get_field_name(
            value=field, fields_map=fields_map, field_manual=field_manual
        )

        inner = f'{field} == "{value}"'

        kwargs["query"] = self._build_query(
            inner=inner, pre=pre, post=post, not_flag=not_flag,
        )

        return self.get(fields_map=fields_map, **kwargs)

    def _get_page(self, state, store):
        page_start = dt_now()
        state["page_num"] += 1

        if state["use_cursor"]:
            rows = self._get_page_cursor(state=state, store=store)
        else:
            rows = self._get_page_normal(state=state, store=store)

        self._get_page_handle(state=state, store=store, rows=rows)

        time_page = dt_sec_ago(obj=page_start)
        rows_page = len(rows)

        state["rows_start"] += rows_page
        state["rows_fetched"] += rows_page
        state["rows_page"] = rows_page
        state["time_page"] = time_page
        state["time_fetch"] += time_page

        self.LOG.debug(f"FETCHED PAGE: {state}")
        return rows

    def _get_page_cursor(self, state, store):
        state["page"] = self._get_cursor(
            query=store["query"],
            fields=store["fields"],
            page_size=state["page_size"],
            cursor=state["page_cursor"],
        )
        rows = state["page"].pop("assets", [])
        state["page_cursor"] = state["page"].get("cursor")
        return rows

    def _get_page_normal(self, state, store):
        state["page"] = self._get(
            query=store["query"],
            fields=store["fields"],
            row_start=state["rows_start"],
            page_size=state["page_size"],
        )
        rows = state["page"].pop("assets", [])
        return rows

    def _build_query(self, inner, not_flag=False, pre="", post=""):
        """Pass."""
        if not_flag:
            inner = f"(not ({inner}))"
        else:
            inner = f"({inner})"

        lines = [pre, inner, post]
        query = " ".join([x.strip() for x in lines if x.strip()]).strip()

        self.LOG.debug(f"Built query: {query!r}")
        # XXX error if no OR / AND in pre & post
        return query

    def _init(self, auth, **kwargs):
        """Post init method for subclasses to use for extra setup.

        Args:
            auth (:obj:`.auth.Model`): object to use for auth and sending API requests
        """
        # cross reference
        self.adapters = Adapters(auth=auth, **kwargs)

        # children
        self.labels = Labels(parent=self)
        self.saved_query = SavedQuery(parent=self)
        self.fields = Fields(parent=self)

        super(AssetMixin, self)._init(auth=auth, **kwargs)

    def _count(self, query=None):
        """Direct API method to get the count of assets.

        Args:
            query (:obj:`str`, optional): default ``None`` -

                * if ``None`` return the count of all assets
                * if :obj:`str` return the count of assets that match a
                  query built by the GUI query wizard

        Returns:
            :obj:`int`: count of assets matching query
        """
        params = {}
        if query:
            params["filter"] = query

        return self.request(method="post", path=self.router.count, json=params)

    def _get(self, query=None, fields=None, row_start=0, page_size=0):
        """Direct API method to get a page of assets.

        Args:
            query (:obj:`str`, optional): default ``None`` -

                * if ``None`` return all assets
                * if :obj:`str` return the assets that match a query built
                  by the GUI query wizard
            fields (:obj:`list` of :obj:`str` or :obj:`str`): default ``None`` -

                * if :obj:`str` CSV seperated list of fields (columns) to include in
                  return
                * if :obj:`list` of :obj:`str` the strs of fields (columns) to include
                  in return
            row_start (:obj:`int`, optional): default ``0`` - for paging, skip N rows
            page_size (:obj:`int`, optional): default ``0`` - for paging, return N rows

        Returns:
            :obj:`list` of :obj:`dict`: assets matching **query** with key/value pairs
                requested as per **fields**
        """
        page_size = self._get_page_size(page_size=page_size, max_rows=None)

        params = {}
        params["skip"] = row_start
        params["limit"] = page_size

        if query:
            params["filter"] = query

        if fields:
            if isinstance(fields, list):
                fields = ",".join(fields)

            params["fields"] = fields

        self._LAST_GET = params

        return self.request(method="post", path=self.router.root, json=params)

    def _get_cursor(self, query=None, fields=None, page_size=PAGE_SIZE, cursor=None):
        """Get a page for a given query."""
        page_size = self._get_page_size(page_size=page_size, max_rows=None)

        params = {}
        params["cursor"] = cursor
        params["limit"] = page_size

        if query:
            params["filter"] = query

        if fields:
            if isinstance(fields, list):
                fields = ",".join(fields)

            params["fields"] = fields

        self._LAST_GET = params

        return self.request(method="post", path=self.router.cached, json=params)

    def _get_by_id(self, id):
        """Direct API method to get the full metadata of all adapters for a single asset.

        Args:
            id (:obj:`str`): internal_axon_id of asset to get all metadata for

        Returns:
            :obj:`dict`: dict with all metadata for all adapters for asset with
                **id** of internal_axon_id
        """
        path = self.router.by_id.format(id=id)
        return self.request(method="get", path=path)

    def _destroy(self, destroy, history):
        """Destroy ALL assets."""
        data = {"destroy": destroy, "history": history}
        path = self.router.destroy
        return self.request(method="post", path=path, json=data)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...context import AliasedGroup
from . import (cmd_add, cmd_delete, cmd_get, cmd_get_default, cmd_set_default,
               cmd_update)


@click.group(cls=AliasedGroup)
def roles():
    """Group: Manage Roles."""


roles.add_command(cmd_get.cmd)
roles.add_command(cmd_get_default.cmd)
roles.add_command(cmd_set_default.cmd)
roles.add_command(cmd_delete.cmd)
roles.add_command(cmd_add.cmd)
roles.add_command(cmd_update.cmd)



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
import codecs
import csv

from ...tools import listify
from .base import Base


class Csv(Base):
    """Pass."""

    CB_NAME = "csv"

    def row(self, row):
        """Write row to dictwriter and delete it."""
        row_return = [{"internal_axon_id": row["internal_axon_id"]}]

        rows = super(Csv, self).row(row=row)

        for row_new in listify(rows):
            self._stream.writerow(row_new)
            del row_new

        del rows
        del row

        return row_return

    def start(self, **kwargs):
        """Create csvstream and associated file descriptor."""
        super(Csv, self).start(**kwargs)
        self.GETARGS["field_null"] = True
        self.GETARGS["field_flatten"] = True
        self.GETARGS["field_join"] = True
        self.GETARGS["field_titles"] = True

        restval = self.GETARGS.get("csv_key_miss", None)
        dialect = self.get_csv_dialect()
        quote = self.get_csv_quote()

        final = self.schemas_final(flat=True)
        titles = [x["column_title"] for x in final]

        self.open_fd()
        self._fd.write(codecs.BOM_UTF8.decode("utf-8"))
        self._stream = csv.DictWriter(
            self._fd,
            fieldnames=titles,
            quoting=quote,
            lineterminator="\n",
            restval=restval,
            dialect=dialect,
        )
        self._stream.writerow(dict(zip(titles, titles)))
        self.do_export_schema(final=final)

    def get_csv_dialect(self):
        """Pass."""
        dialect = self.GETARGS.get("csv_dialect", "excel")
        return dialect

    def get_csv_quote(self):
        """Pass."""
        quote = self.GETARGS.get("csv_quoting", "nonnumeric")
        quote = getattr(csv, f"QUOTE_{quote.upper()}")
        return quote

    def do_export_schema(self, final):
        """Pass."""
        export_schema = self.GETARGS.get("export_schema", True)

        if export_schema:
            titles = [x["column_title"] for x in final]
            names = [x["name_qual"] for x in final]
            types = [x["type_norm"] for x in final]

            self._stream.writerow(dict(zip(titles, names)))
            self._stream.writerow(dict(zip(titles, types)))

    def stop(self, **kwargs):
        """Close dictwriter and associated file descriptor."""
        super(Csv, self).stop(**kwargs)
        self._fd.write("\n")
        self.close_fd()



# -*- coding: utf-8 -*-
"""Test suite."""
import pytest


class SystemDiscoverBase:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_system):
        """Pass."""
        return api_system.discover


class TestSystemDiscoverPrivate(SystemDiscoverBase):
    """Pass."""

    def test_private_lifecycle(self, apiobj):
        """Pass."""
        lifecycle = apiobj._get()
        assert isinstance(lifecycle, dict)
        assert "status" in lifecycle
        assert lifecycle["status"] in ["starting", "running", "done"]

    def test_private_start_stop(self, apiobj):
        """Pass."""
        stop = apiobj._stop()
        assert not stop

        lifecycle = apiobj._get()
        assert lifecycle["status"] in ["done"]

        start = apiobj._start()
        assert not start

        lifecycle = apiobj._get()
        assert lifecycle["status"] in ["starting", "running"]

        re_stop = apiobj._stop()
        assert not re_stop

        lifecycle = apiobj._get()
        assert lifecycle["status"] in ["done"]


class TestSystemDiscoverPublic(SystemDiscoverBase):
    """Pass."""

    def test_get(self, apiobj):
        """Pass."""
        lifecycle = apiobj.get()
        assert isinstance(lifecycle, dict)
        assert isinstance(lifecycle["is_running"], bool)

    def test_start_stop(self, apiobj):
        """Pass."""
        if apiobj.is_running:
            stopped = apiobj.stop()
            assert isinstance(stopped, dict)
            assert not stopped["is_running"]
            # assert not stopped["status"] == "done"

        started = apiobj.start()
        assert isinstance(started, dict)
        assert started["is_running"]
        # assert started["status"] in ["starting", "running"]

        re_stopped = apiobj.stop()
        assert isinstance(re_stopped, dict)
        assert not re_stopped["is_running"]
        # assert re_stopped["status"] == "done"



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...context import AliasedGroup
from . import cmd_add, cmd_delete, cmd_get, cmd_get_by_name, cmd_update


@click.group(cls=AliasedGroup)
def users():
    """Group: Manage Users."""


users.add_command(cmd_get.cmd)
users.add_command(cmd_get_by_name.cmd)
users.add_command(cmd_update.cmd)
users.add_command(cmd_add.cmd)
users.add_command(cmd_delete.cmd)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import tabulate

from ...constants import (FIELD_JOINER, FIELD_TRIM_LEN, TABLE_FORMAT,
                          TABLE_MAX_ROWS)
from ..context import CONTEXT_SETTINGS, click
from ..options import (AUTH, EXPORT, FIELDS_SELECT, PAGING, add_options,
                       get_option_fields_default, get_option_help)

GET_EXPORT = [
    click.option(
        "--use-cursor/--no-use-cursor",
        "use_cursor",
        default=True,
        help="Use cursor for pagination",
        hidden=True,
    ),
    click.option(
        "--echo/--no-echo",
        "do_echo",
        default=True,
        help="Print out details during fetch",
        show_envvar=True,
        show_default=True,
        is_flag=True,
        hidden=False,
    ),
    click.option(
        "--export-format",
        "-xt",
        "export",
        default="json",
        help="Formatter to use when exporting asset data",
        type=click.Choice(["csv", "json", "table"]),
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--table-format",
        "table_format",
        default=TABLE_FORMAT,
        help="Table format to use for --export-format=table",
        type=click.Choice(tabulate.tabulate_formats),
        show_envvar=True,
        show_default=True,
        hidden=False,
    ),
    click.option(
        "--table-max-rows",
        "table_max_rows",
        default=TABLE_MAX_ROWS,
        help="Only return this many rows for --export-format=table",
        show_envvar=True,
        show_default=True,
        type=click.INT,
        hidden=False,
    ),
    click.option(
        "--schema/--no-schema",
        "export_schema",
        default=False,
        help="Add schema information to the export",
        is_flag=True,
        show_envvar=True,
        show_default=True,
        hidden=False,
    ),
    click.option(
        "--titles/--no-titles",
        "field_titles",
        default=False,
        help="Rename fields from internal field names to their column titles",
        is_flag=True,
        show_envvar=True,
        show_default=True,
        hidden=False,
    ),
    click.option(
        "--join/--no-join",
        "field_join",
        default=False,
        help="Join multivalue fields using --join-value",
        is_flag=True,
        show_envvar=True,
        show_default=True,
        hidden=False,
    ),
    click.option(
        "--join-value",
        "field_join_value",
        default=FIELD_JOINER,
        help="Value to use for joining multivalue fields, default: \\n",
        show_envvar=True,
        show_default=True,
        hidden=False,
    ),
    click.option(
        "--join-trim",
        "field_join_trim",
        default=FIELD_TRIM_LEN,
        help="Character length to trim joined multivalue fields",
        show_envvar=True,
        show_default=True,
        type=click.INT,
        hidden=False,
    ),
    click.option(
        "--explode",
        "field_explode",
        default="",
        help="Flatten and explode a fields values into multiple rows",
        show_envvar=True,
        show_default=True,
        hidden=False,
        metavar="FIELD",
    ),
    click.option(
        "--flatten/--no-flatten",
        "field_flatten",
        default=False,
        help="Remove complex fields and re-add their sub-field values to the row",
        show_envvar=True,
        show_default=True,
        is_flag=True,
        hidden=False,
    ),
    click.option(
        "--field-exclude",
        "-fx",
        "field_excludes",
        help="Fields to exclude from each row (multiples)",
        multiple=True,
        default=[],
        show_envvar=True,
        show_default=True,
        hidden=False,
        metavar="FIELD",
    ),
    click.option(
        "--field-null/--no-field-null",
        "field_null",
        default=False,
        help="Add missing fields with --field-null-value",
        show_envvar=True,
        show_default=True,
        is_flag=True,
        hidden=False,
    ),
    click.option(
        "--field-null-value",
        "field_null_value",
        default=None,
        help="Value to use for fields that are not returned",
        show_envvar=True,
        show_default=True,
        hidden=False,
    ),
    click.option(
        "--adapters-missing/--no-adapters-missing",
        "report_adapters_missing",
        default=False,
        help="Add a column showing adapters missing from each asset",
        show_envvar=True,
        show_default=True,
        is_flag=True,
        hidden=False,
    ),
    click.option(
        "--tag",
        "tags_add",
        help="Tags to add to each asset (multiples)",
        multiple=True,
        default=[],
        show_envvar=True,
        show_default=True,
        hidden=False,
        metavar="TAG",
    ),
    click.option(
        "--untag",
        "tags_remove",
        help="Tags to remove from each asset (multiples)",
        multiple=True,
        default=[],
        show_envvar=True,
        show_default=True,
        hidden=False,
        metavar="TAG",
    ),
]

GET_BUILDERS = [
    *AUTH,
    *PAGING,
    *EXPORT,
    *GET_EXPORT,
    *FIELDS_SELECT,
    get_option_fields_default(default=True),
    click.option(
        "--pre",
        "pre",
        help="Query to prefix the query built to search for --value",
        default="",
        metavar="QUERY",
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--post",
        "post",
        help="Query to postfix the query built to search for --value",
        default="",
        metavar="QUERY",
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--not-flag/--no-not-flag",
        "-nf/-nnf",
        "not_flag",
        help="Build the query as a NOT query",
        is_flag=True,
        default=False,
        show_envvar=True,
        show_default=True,
    ),
]

GET_BY_VALUE_BUILDERS = [
    *GET_BUILDERS,
    click.option(
        "--value",
        "-v",
        "value",
        help="Value to search for",
        required=True,
        show_envvar=True,
        show_default=True,
    ),
    get_option_help(choices=["auth", "assetexport", "selectfields"]),
]

GET_BY_VALUES_BUILDERS = [
    *GET_BUILDERS,
    click.option(
        "--value",
        "-v",
        "values",
        help="Values to search for (multiples)",
        required=True,
        multiple=True,
        show_envvar=True,
        show_default=True,
    ),
    get_option_help(choices=["auth", "assetexport", "selectfields"]),
]

GET_BY_VALUE_REGEX_BUILDERS = [
    *GET_BUILDERS,
    click.option(
        "--value",
        "-v",
        "value",
        help="Regular expression value to search for",
        required=True,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--case-in-sensitive/--case-sensitive",
        "-cis/-cs",
        "cast_insensitive",
        help="Build a regex that is / is not cases sensitive",
        is_flag=True,
        default=True,
        show_envvar=True,
        show_default=True,
    ),
    get_option_help(choices=["auth", "assetexport", "selectfields"]),
]

GET_BY_VALUE_FIELD = click.option(
    "--field",
    "-f",
    "field",
    help="Field to use when building the query",
    required=True,
    show_envvar=True,
    show_default=True,
)


def gen_get_by_cmd(options, doc, cmd_name, method):
    """Pass."""

    @click.command(name=cmd_name, context_settings=CONTEXT_SETTINGS, help=doc)
    @add_options(options)
    @click.pass_context
    def cmd(ctx, url, key, secret, get_method=method, **kwargs):
        client = ctx.obj.start_client(url=url, key=key, secret=secret)

        p_grp = ctx.parent.command.name
        apiobj = getattr(client, p_grp)
        apimethod = getattr(apiobj, get_method)

        with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
            apimethod(**kwargs)

    cmd.__doc__ = doc
    return cmd



# -*- coding: utf-8 -*-
"""API models for working with adapters and connections."""
import copy
import textwrap

import tabulate

from ...constants import KEY_MAP_ADAPTER, KEY_MAP_CNX, KEY_MAP_SCHEMA
from ...tools import json_dump


def tablize(value, err=None, fmt="simple", footer=True, **kwargs):
    """Pass."""
    # value = wrapper(value=value, **kwargs)

    table = tabulate.tabulate(value, tablefmt=fmt, headers="keys")

    if footer:
        if fmt == "simple":
            header = "\n" + "\n".join(reversed(table.splitlines()[0:2]))
            table += header

    if err:
        pre = err + "\n"
        post = "\n" + err
    else:
        pre = "\n"
        post = "\n"

    return "\n".join([pre, table, post])


def tablize_schemas(
    schemas, config=None, err=None, fmt="simple", footer=True, orig=True, orig_width=20
):
    """Pass."""
    values = []
    config = config or None
    if isinstance(schemas, dict):
        schemas = list(schemas.values())

    for schema in sorted(schemas, key=lambda x: [x["required"], x["name"]]):
        value = tab_map(
            value=schema, key_map=KEY_MAP_SCHEMA, orig=orig, orig_width=orig_width
        )
        if config:
            config_value = config.get(schema["name"], None)
            if isinstance(config_value, dict):
                config_value = json_dump(config_value)
            value["Current Value"] = config_value
        values.append(value)

    return tablize(value=values, err=err, fmt=fmt, footer=footer)


def tablize_adapters(adapters, err=None, fmt="simple", footer=True):
    """Pass."""
    values = []

    for adapter in adapters:
        value = tab_map(value=adapter, key_map=KEY_MAP_ADAPTER, orig=False)
        # value["Connection IDs"] = "\n".join([x["id"] for x in adapter["cnx"]])
        values.append(value)

    return tablize(value=values, err=err, fmt=fmt, footer=footer)


def tablize_cnxs(cnxs, err=None, fmt="simple", footer=True):
    """Pass."""
    values = []
    for cnx in cnxs:
        cnx["label"] = cnx["config"].get("connection_label")
        value = tab_map(value=cnx, key_map=KEY_MAP_CNX, orig=False)
        values.append(value)

    return tablize(value=values, err=err, fmt=fmt, footer=footer)


def tab_map(value, key_map, orig=False, orig_width=20):
    """Pass."""
    orig_value = copy.deepcopy(value)

    new_value = {}

    for key, name, width in key_map:
        if key in orig_value and name:

            key_value = orig_value.pop(key)

            if isinstance(key_value, list):
                key_value = "\n".join([str(x) for x in key_value])

            if isinstance(key_value, str) and key_value and width:
                key_value = textwrap.fill(key_value, width=width)

            new_value[name] = key_value

    if orig:
        for orig_key, orig_key_value in orig_value.items():
            if isinstance(orig_key_value, dict):
                continue

            if isinstance(orig_key_value, list):
                new_value[orig_key] = "\n".join([str(x) for x in orig_key_value])

            elif isinstance(orig_key_value, str) and orig_key_value and orig_width:
                new_value[orig_key] = textwrap.fill(orig_key_value, width=orig_width)
            else:
                new_value[orig_key] = orig_key_value

    return new_value



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...context import AliasedGroup
from . import cmd_get, cmd_start, cmd_stop


@click.group(cls=AliasedGroup)
def discover():
    """Group: Discover and Lifecycle management."""


discover.add_command(cmd_get.cmd)
discover.add_command(cmd_start.cmd)
discover.add_command(cmd_stop.cmd)



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
import pytest

from axonius_api_client.auth import ApiKey
from axonius_api_client.exceptions import (
    AlreadyLoggedIn,
    AuthError,
    InvalidCredentials,
    NotLoggedIn,
)
from axonius_api_client.http import Http

from ..utils import get_key_creds, get_url


class TestApiKey:
    """Test axonius_api_client.auth."""

    def test_valid_creds(self, request):
        """Test str/repr has URL."""
        http = Http(url=get_url(request), certwarn=False)

        auth = ApiKey(http=http, **get_key_creds(request))

        auth.login()

        assert auth.is_logged_in
        assert "url" in format(auth)
        assert "url" in repr(auth)

    def test_logout(self, request):
        """Test no exc when logout() after login()."""
        http = Http(url=get_url(request), certwarn=False)

        auth = ApiKey(http=http, **get_key_creds(request))

        auth.login()

        assert auth.is_logged_in

        auth.logout()

        assert not auth.is_logged_in

    def test_login_already_logged_in(self, request):
        """Test exc thrown when login() and login() already called."""
        http = Http(url=get_url(request), certwarn=False)

        auth = ApiKey(http=http, **get_key_creds(request))

        auth.login()

        with pytest.raises(AlreadyLoggedIn):
            auth.login()

    def test_logout_not_logged_in(self, request):
        """Test exc thrown when logout() but login() not called."""
        http = Http(url=get_url(request), certwarn=False)

        auth = ApiKey(http=http, **get_key_creds(request))

        with pytest.raises(NotLoggedIn):
            auth.logout()

    def test_invalid_creds(self, request):
        """Test str/repr has URL."""
        http = Http(url=get_url(request), certwarn=False)

        bad = "badwolf"

        auth = ApiKey(http=http, key=bad, secret=bad)

        with pytest.raises(InvalidCredentials):
            auth.login()

    def test_http_lock_fail(self, request):
        """Test using an http client from another authmethod throws exc."""
        http = Http(url=get_url(request), certwarn=False)

        auth = ApiKey(http=http, **get_key_creds(request))

        assert auth.http._auth_lock

        with pytest.raises(AuthError):
            auth = ApiKey(http=http, **get_key_creds(request))



# -*- coding: utf-8 -*-
"""API models for working with adapters and connections."""
from ...constants import DEFAULT_NODE
from ...exceptions import ApiError, NotFoundError
from ...tools import path_read
from ..mixins import ModelMixins
from ..parsers.adapters import parse_adapters, parse_schema
from ..parsers.config import config_build, config_unchanged, config_unknown
from ..parsers.tables import tablize_adapters
from ..routers import API_VERSION
from .cnx import Cnx


class Adapters(ModelMixins):
    """API model for working with adapters."""

    @property
    def router(self):
        """Router for this API model."""
        return API_VERSION.adapters

    def get(self):
        """Pass."""
        parsed = parse_adapters(raw=self._get())
        parsed = sorted(parsed, key=lambda x: [x["node_name"], x["name"]])
        return parsed

    def get_by_name(self, name, node=DEFAULT_NODE):
        """Pass."""
        adapters = self.get()

        keys = ["node_name", "node_id"]
        nodes = [x for x in adapters if node.lower() in [x[k].lower() for k in keys]]

        if not nodes:
            err = f"No node named {node!r} found"
            raise NotFoundError(tablize_adapters(adapters=adapters, err=err))

        keys = ["name", "name_raw", "name_plugin"]
        for adapter in nodes:
            if any([adapter[k].lower() == name.lower() for k in keys]):
                return adapter

        err = f"No adapter named {name!r} found on node {node!r}"
        raise NotFoundError(tablize_adapters(adapters=adapters, err=err))

    def config_get(self, name, node=DEFAULT_NODE, generic=True):
        """Pass."""
        adapter = self.get_by_name(name=name, node=node)
        return self.config_refetch(adapter=adapter, generic=generic)

    def config_refetch(self, adapter, generic=True):
        """Pass."""
        name_config = "generic_name" if generic else "specific_name"
        name_config = adapter["schemas"][name_config]
        name_plugin = adapter["name_plugin"]

        if not generic and not name_config:
            name = adapter["name"]
            raise ApiError(f"Adapter {name} has no adapter specific advanced settings!")

        data = self._config_get(name_plugin=name_plugin, name_config=name_config)

        data["schema"] = parse_schema(raw=data["schema"])

        return data

    def config_update(self, name, node=DEFAULT_NODE, generic=True, **kwargs):
        """Pass."""
        kwargs_config = kwargs.pop("kwargs_config", {})
        kwargs.update(kwargs_config)
        adapter = self.get_by_name(name=name, node=node)

        name_config = "generic_name" if generic else "specific_name"
        name_config_src = "generic" if generic else "adapter specific"
        name_config = adapter["schemas"][name_config]

        config_map = self.config_refetch(adapter=adapter, generic=generic)
        old_config = config_map["config"]
        schemas = config_map["schema"]

        source = f"adapter {name!r} {name_config_src} advanced settings"
        config_unknown(schemas=schemas, new_config=kwargs, source=source)
        new_config = config_build(
            schemas=schemas, old_config=old_config, new_config=kwargs, source=source
        )
        config_unchanged(
            schemas=schemas, old_config=old_config, new_config=new_config, source=source
        )

        self._config_update(
            name_raw=adapter["name_raw"], name_config=name_config, new_config=new_config,
        )

        return self.config_refetch(adapter=adapter, generic=generic)

    def file_upload(
        self,
        name,
        field_name,
        file_name,
        file_content,
        file_content_type=None,
        node=DEFAULT_NODE,
    ):
        """Upload a str as a file to a specific adapter on a specific node.

        Args:
            adapter (:obj:`str` or :obj:`dict`):

                * If :obj:`str`, the name of the adapter to get the metadata for
                * If :obj:`dict`, the metadata for a previously fetched adapter
            field (:obj:`str`): name of field to store data in
            name (:obj:`str`): filename to use when uploading file
            content (:obj:`str` or :obj:`bytes`): content of file to upload
            node (:obj:`str`, optional): default
                :data:`.DEFAULT_NODE` -
                node name running ``adapter``
            content_type (:obj:`str`, optional): default ``None`` -
                mime type of **content**

        Returns:
            :obj:`dict`: dict that can be used in a connection parameter of type
            file that contains the keys:

              * uuid: UUID of uploaded file
              * filename: filename of uploaded file
        """
        adapter = self.get_by_name(name=name, node=node)

        return self._file_upload(
            name_raw=adapter["name_raw"],
            node_id=adapter["node_id"],
            file_name=file_name,
            field_name=field_name,
            file_content=file_content,
            file_content_type=file_content_type,
        )

    def file_upload_path(self, path, **kwargs):
        """Upload the contents of a file to a specific adapter on a specific node.

        Args:
            adapter (:obj:`str` or :obj:`dict`):

                * If :obj:`str`, the name of the adapter to get the metadata for
                * If :obj:`dict`, the metadata for a previously fetched adapter
            field (:obj:`str`): name of field to store data in
            name (:obj:`str`): filename to use when uploading file
            path (:obj:`str` or :obj:`pathlib.Path`): path to file containing contents
                to upload
            node (:obj:`str`, optional): default
                :data:`.DEFAULT_NODE` -
                node name running ``adapter``
            content_type (:obj:`str`, optional): default ``None`` -
                mime type of **content**

        Returns:
            :obj:`dict`: dict that can be used in a connection parameter of type
            file that contains the keys:

              * uuid: UUID of uploaded file
              * filename: filename of uploaded file
        """
        path, file_content = path_read(obj=path, binary=True, is_json=False)
        kwargs.setdefault("field_name", path.name)
        kwargs.setdefault("file_name", path.name)
        kwargs["file_content"] = file_content
        return self.file_upload(**kwargs)

    def _init(self, auth, **kwargs):
        """Post init method for subclasses to use for extra setup.

        Args:
            auth (:obj:`.auth.Model`): object to use for auth and sending API requests
        """
        self.cnx = Cnx(parent=self)
        super(Adapters, self)._init(auth=auth, **kwargs)

    def _get(self):
        """Direct API method to get all adapters.

        Returns:
            :obj:`dict`: raw metadata for all adapters
        """
        path = self.router.root
        return self.request(method="get", path=path)

    def _config_update(self, name_raw, name_config, new_config):
        """Direct API method to set advanced settings for an adapter.

        Args:
            name_raw (:obj:`str`): unique plugin name of adapter to set
                advanced settings for. unique plugin name refers to the name of
                an adapter for a specific node
            name_config (:obj:`str`): name of advanced settings to set -
                either AdapterBase for generic advanced settings or
                AwsSettings for adapter specific settings, if applicable
            config (:obj:`dict`): the configuration values to set

        Returns:
            :obj:`str`: empty str
        """
        path = self.router.config_set.format(
            adapter_name_raw=name_raw, adapter_config_name=name_config
        )
        return self.request(
            method="post", path=path, json=new_config, error_json_invalid=False
        )

    def _config_get(self, name_plugin, name_config):
        """Direct API method to set advanced settings for an adapter.

        Args:
            name_plugin (:obj:`str`): unique plugin name of adapter to set
                advanced settings for. unique plugin name refers to the name of
                an adapter for a specific node i.e. aws_adapter_0
            name_config (:obj:`str`): name of advanced settings to set -
                either AdapterBase for generic advanced settings or
                AwsSettings for adapter specific settings, if applicable

        Returns:
            :obj:`dict`: current settings for name_config of name_plugin
        """
        path = self.router.config_get.format(
            adapter_name_plugin=name_plugin, adapter_config_name=name_config
        )
        return self.request(method="get", path=path)

    def _file_upload(
        self,
        name_raw,
        node_id,
        field_name,
        file_name,
        file_content,
        file_content_type=None,
        file_headers=None,
    ):
        """Direct API method to upload a file to a specific adapter on a specifc node.

        Args:
            name_raw (:obj:`str`): name of the adapter to upload a file to
            node_id (:obj:`str`): ID of node running **adapter**
            field_name (:obj:`str`): name of setting this file is intended for
            file_name (:obj:`str`): filename to use when uploading file
            file_content (:obj:`str` or :obj:`bytes`): content to upload
            file_content_type (:obj:`str`, optional): default ``None`` -
                mime type of **content**
            file_headers (:obj:`dict`, optional): default ``None`` -
                headers to use for file **content**

        Returns:
            :obj:`dict`: dict that can be used in a connection parameter of type
            file that contains the keys:

              * uuid: UUID of uploaded file
              * filename: filename of uploaded file
        """
        data = {"field_name": field_name}
        files = {"userfile": (file_name, file_content, file_content_type, file_headers)}
        path = self.router.file_upload.format(
            adapter_name_raw=name_raw, adapter_node_id=node_id
        )
        ret = self.request(method="post", path=path, data=data, files=files)
        ret["filename"] = file_name
        return ret

    # def _download_file(self, name, node_id, cnx_id, schema_key):
    #     """Direct API method to download a file.

    #     Notes:
    #         WORK IN PROGRESS!

    #     """
    #     data = {"uuid": cnx_id, "schema_key": schema_key}
    #     path = self.router.download_file.format(
    #         name=name, node_id=node_id
    #     )
    #     ret = self.request(method="post", path=path, json=data, raw=True)
    #     return ret



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...context import AliasedGroup
from . import cmd_get


@click.group(cls=AliasedGroup)
def nodes():
    """Group: Manage Nodes (Instances)."""


nodes.add_command(cmd_get.cmd)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...context import AliasedGroup
from . import cmd_about, cmd_sizes


@click.group(cls=AliasedGroup)
def meta():
    """Group: System metadata."""


meta.add_command(cmd_about.cmd)
meta.add_command(cmd_sizes.cmd)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....api.parsers.tables import tablize_cnxs, tablize_schemas
from ....constants import CNX_SANE_DEFAULTS
from ....exceptions import CnxAddError
from ....tools import json_dump, listify, pathlib
from ...context import click

PATH_PROMPT = click.Path(
    exists=True,
    file_okay=True,
    dir_okay=False,
    writable=False,
    readable=True,
    resolve_path=True,
    allow_dash=False,
    path_type=None,
)

HIDDEN = ["secret", "key", "password"]

PROMPTS = [
    click.option(
        "--prompt-optional / --no-prompt-optional",
        "-po / -npo",
        "prompt_optional",
        help="Prompt for optional items that are not supplied.",
        is_flag=True,
        default=True,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--prompt-default / --no-prompt-default",
        "-pd / -npd",
        "prompt_default",
        help="Prompt for items that have a default.",
        is_flag=True,
        default=True,
        show_envvar=True,
        show_default=True,
    ),
]

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(
        ["json-full", "json-config", "json", "table", "table-schemas", "str", "str-args"]
    ),
    help="Format of to export data in",
    default="json-config",
    show_envvar=True,
    show_default=True,
)
ID_CNX = click.option(
    "--id",
    "-i",
    "cnx_id",
    help="ID of connection",
    required=True,
    show_envvar=True,
    show_default=True,
)


def prompt_config(
    ctx,
    client,
    new_config,
    adapter_name,
    adapter_node,
    prompt_optional,
    prompt_default,
    **kwargs,
):
    """Pass."""
    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        adapter = client.adapters.get_by_name(name=adapter_name, node=adapter_node)

    schemas = list(adapter["schemas"]["cnx"].values())
    schemas = reversed(sorted(schemas, key=lambda x: [x["required"], x["name"]]))

    for schema in schemas:
        prompt_schema(
            schema=schema,
            new_config=new_config,
            prompt_optional=prompt_optional,
            prompt_default=prompt_default,
            adapter=adapter,
        )


def prompt_schema(schema, new_config, prompt_optional, prompt_default, adapter):
    """Pass."""
    name = schema["name"]
    required = schema["required"]
    stype = schema["type"]
    default = schema.get("default", None)
    fmt = schema.get("format", "")
    enum = schema.get("enum", [])

    if enum:
        str_type = click.Choice(choices=enum, case_sensitive=True)
    else:
        str_type = None

    schema["hide_input"] = hide_input = fmt == "password" or name in HIDDEN

    if name in new_config:
        if new_config[name].lower().strip() == "_empty_":
            new_config[name] = None
        if stype == "file":
            new_config[name] = pathlib.Path(new_config[name]).expanduser().resolve()
        return

    sane_defaults = CNX_SANE_DEFAULTS.get(adapter["name"], CNX_SANE_DEFAULTS["all"])
    if name in sane_defaults and default is None:
        default = sane_defaults[name]

    if not prompt_default and default is not None:
        new_config[name] = default
        return

    if not required:
        if not prompt_optional:
            return

        show_schema(schema=schema)

        do_prompt = click.prompt(
            text="Prompt for this optional item?",
            default=True,
            type=click.BOOL,
            err=True,
            show_default=True,
        )

        if not do_prompt:
            return
    else:
        show_schema(schema=schema)

    if stype == "file":
        value = click.prompt(text="Enter value", type=PATH_PROMPT, err=True)
        value = pathlib.Path(value).expanduser().resolve()

    if stype == "bool":
        value = click.prompt(
            text="Enter value",
            default=default,
            type=click.BOOL,
            err=True,
            show_default=True,
            show_choices=True,
        )

    if stype in ["number", "integer"]:
        value = click.prompt(
            text="Enter value",
            default=default,
            type=click.INT,
            err=True,
            show_default=True,
            show_choices=True,
        )

    if stype == "array":
        value = click.prompt(
            text="Enter value",
            default=default,
            err=True,
            type=str_type,
            hide_input=hide_input,
            show_default=True,
            show_choices=True,
        )
        value = [x.strip() for x in value.strip().split(",") if x.strip()]

    if stype == "string":
        value = click.prompt(
            text="Enter value",
            default=default,
            err=True,
            type=str_type,
            hide_input=hide_input,
            show_default=True,
            show_choices=True,
        )

    new_config[name] = value


def show_schema(schema, err=True):
    """Pass."""
    rkw = ["{}: {}".format(k, v) for k, v in schema.items()]
    rkw = "\n  " + "\n  ".join(rkw)
    click.secho(message=f"\n***  Configuration schema:{rkw}", fg="blue", err=err)


def add_cnx(ctx, client, adapter_name, adapter_node, new_config, **kwargs):
    """Pass."""
    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        try:
            cnx_new = client.adapters.cnx.add(
                adapter_name=adapter_name, adapter_node=adapter_node, **new_config,
            )
            ctx.obj.echo_ok(msg=f"Connection added successfully!")

        except CnxAddError as exc:
            ctx.obj.echo_error(msg=f"{exc}", abort=False)
            cnx_new = exc.cnx_new

    handle_export(ctx=ctx, rows=cnx_new, **kwargs)


def handle_export(ctx, rows, export_format, **kwargs):
    """Pass."""
    if export_format == "json":
        if isinstance(rows, list):
            for row in rows:
                row.pop("schemas")
        else:
            rows.pop("schemas")
        click.secho(json_dump(rows))
        ctx.exit(0)

    if export_format == "json-config":
        if isinstance(rows, list):
            rows = [x["config"] for x in rows]
        else:
            rows = rows["config"]
        click.secho(json_dump(rows))
        ctx.exit(0)

    if export_format == "json-full":
        click.secho(json_dump(rows))
        ctx.exit(0)

    if export_format == "table":
        rows = listify(rows)
        click.secho(tablize_cnxs(cnxs=rows))
        ctx.exit(0)

    if export_format == "table-schemas":
        if isinstance(rows, list):
            schemas = rows[0]["schemas"]
        else:
            schemas = rows["schemas"]
        click.secho(
            tablize_schemas(
                schemas=schemas,
                err=None,
                fmt="simple",
                footer=True,
                orig=True,
                orig_width=20,
            )
        )
        ctx.exit(0)

    if export_format == "str-args":
        rows = listify(rows)
        lines = "\n".join(
            [
                "--node-name {node_name} --name {adapter_name} --id {id}".format(**row)
                for row in rows
            ]
        )
        click.secho(lines)
        ctx.exit(0)

    if export_format == "str":
        rows = listify(rows)
        lines = "\n".join(["{id}".format(**row) for row in rows])
        click.secho(lines)
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""Exceptions and warnings."""


class AxonWarning(Warning):
    """Warnings for :mod:`axonius_api_client`."""


class ApiWarning(AxonWarning):
    """Warnings for :mod:`.api`."""


class AxonError(Exception):
    """Errors for :mod:`axonius_api_client`."""


class ApiError(AxonError):
    """Errors for :mod:`.api`."""


class ToolsError(AxonError):
    """Errors for :mod:`.tools`."""


class AuthError(AxonError):
    """Errors for :mod:`.auth`."""


class NotFoundError(ApiError):
    """Error when something is not found."""


class InvalidCredentials(AuthError):
    """Error when credentials are invalid."""


class NotLoggedIn(AuthError):
    """Error when not logged in."""


class AlreadyLoggedIn(AuthError):
    """Error when already logged in."""


class ConnectError(AxonError):
    """Error in :obj:`.connect.Connect`."""


class HttpError(AxonError):
    """Errors for :mod:`.http`."""


class ConfigError(ApiError):
    """Pass."""


class ConfigInvalidValue(ConfigError):
    """Pass."""


class ConfigUnchanged(ConfigError):
    """Pass."""


class ConfigUnknown(ConfigError):
    """Pass."""


class ConfigRequired(ConfigError):
    """Pass."""


class CnxError(ApiError):
    """Errors for :obj:`.api.adapters.Cnx`."""


class CnxGoneError(CnxError):
    """Errors for :obj:`.api.adapters.Cnx`."""


class CnxUpdateError(CnxError):
    """Pass."""


class CnxTestError(CnxError):
    """Pass."""


class CnxAddError(CnxError):
    """Pass."""


class ResponseError(ApiError):
    """Errors when checking responses."""


class ResponseNotOk(ResponseError):
    """Error if response has a status code that is an error and error_status is True."""


class JsonInvalid(ResponseError):
    """Error when response has invalid JSON."""


class JsonError(ResponseError):
    """Error when JSON has key:error that is not empty or key:status=error."""



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
from ....cli import cli
from ....constants import CSV_ADAPTER, DEFAULT_NODE
from ....tools import json_load
from ...utils import load_clirunner


class TestGrpCnxCmdGet:
    """Pass."""

    def test_json_full(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get",
            "--node-name",
            DEFAULT_NODE,
            "--name",
            CSV_ADAPTER,
            "--export-format",
            "json-full",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, list)

    def test_json_config(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get",
            "--node-name",
            DEFAULT_NODE,
            "--name",
            CSV_ADAPTER,
            "--export-format",
            "json-config",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, list)

    def test_json_basic(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get",
            "--node-name",
            DEFAULT_NODE,
            "--name",
            CSV_ADAPTER,
            "--export-format",
            "json",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, list)

    def test_table_schemas(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get",
            "--node-name",
            DEFAULT_NODE,
            "--name",
            CSV_ADAPTER,
            "--export-format",
            "table-schemas",
        ]

        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

    def test_table(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get",
            "--node-name",
            DEFAULT_NODE,
            "--name",
            CSV_ADAPTER,
            "--export-format",
            "table",
        ]

        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

    def test_str(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get",
            "--node-name",
            DEFAULT_NODE,
            "--name",
            CSV_ADAPTER,
            "--export-format",
            "str",
        ]

        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

    def test_str_args(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get",
            "--node-name",
            DEFAULT_NODE,
            "--name",
            CSV_ADAPTER,
            "--export-format",
            "str-args",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ..context import CONTEXT_SETTINGS, click
from ..options import (AUTH, EXPORT, FIELDS_SELECT, PAGING, QUERY, add_options,
                       get_option_fields_default, get_option_help)
from .grp_common import GET_EXPORT

OPTIONS = [
    *AUTH,
    *PAGING,
    *EXPORT,
    *GET_EXPORT,
    *FIELDS_SELECT,
    get_option_fields_default(default=True),
    *QUERY,
    get_option_help(choices=["auth", "query", "assetexport", "selectfields"]),
]


@click.command(name="get", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, query_file, **kwargs):
    """Get assets using a query and fields."""
    if query_file:
        kwargs["query"] = query_file.read().strip()

    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        apiobj.get(**kwargs)



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
from ...tools import grouper
from ..mixins import ChildMixins


class Labels(ChildMixins):
    """ChildMixins API model for working with labels/tags for the parent asset type."""

    def add(self, rows, labels):
        """Add labels/tags to assets.

        Args:
            rows (:obj:`list` of :obj:`dict`): assets returned from :meth:`get`
                to process
            labels (:obj:`list` of `str`): labels to process

        Returns:
            :obj:`int`: number of labels processed
        """
        ids = [row["internal_axon_id"] for row in rows]

        processed = 0

        # only do 100 labels at a time, more seems to break API
        for group in grouper(ids, 100):
            group = [x for x in group if x is not None]
            response = self._add(labels=labels, ids=group)
            processed += response

        return processed

    def get(self):
        """Get all known labels/tags.

        Returns:
            :obj:`list` of :obj:`str`: all labels that exist in Axonius
        """
        return self._get()

    def remove(self, rows, labels):
        """Remove labels/tags from assets.

        Args:
            rows (:obj:`list` of :obj:`dict`): assets returned from :meth:`get`
                to process
            labels (:obj:`list` of `str`): labels to process

        Returns:
            :obj:`int`: number of labels processed
        """
        ids = [row["internal_axon_id"] for row in rows]

        processed = 0

        # only do 100 labels at a time, more seems to break API
        for group in grouper(ids, 100):
            group = [x for x in group if x is not None]
            response = self._remove(labels=labels, ids=group)
            processed += response

        return processed

    def _add(self, labels, ids):
        """Direct API method to add labels/tags to assets.

        Args:
            labels (:obj:`list` of `str`): labels to process
            ids (:obj:`list` of `str`): internal_axon_id of assets to add **labels** to

        Returns:
            :obj:`int`: number of labels processed
        """
        data = {}
        data["entities"] = {}
        data["entities"]["ids"] = ids
        data["labels"] = labels

        path = self.router.labels
        return self.request(method="post", path=path, json=data)

    def _get(self):
        """Direct API method to get all known labels/tags.

        Returns:
            :obj:`list` of :obj:`str`: all labels that exist in Axonius
        """
        path = self.router.labels
        return self.request(method="get", path=path)

    def _remove(self, labels, ids):
        """Direct API method to remove labels/tags from assets.

        Args:
            labels (:obj:`list` of `str`): labels to process
            ids (:obj:`list` of `str`): internal_axon_id of assets to remove
                **labels** from

        Returns:
            :obj:`int`: number of labels processed
        """
        data = {}
        data["entities"] = {}
        data["entities"]["ids"] = ids
        data["labels"] = labels

        path = self.router.labels

        return self.request(method="delete", path=path, json=data)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import dt_parse, json_dump
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json", "str"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)

OPTIONS = [
    *AUTH,
    EXPORT,
]


@click.command(name="get", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, **kwargs):
    """Get nodes information."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = client.system.nodes.get()

    if export_format == "json":
        click.secho(json_dump(data))
        ctx.exit(0)

    if export_format == "str":
        for item in data:
            ips = ", ".join(item["ips"])
            last_seen = str(dt_parse(item["last_seen"]))
            name = item["node_name"]
            is_env = item["use_as_environment_name"]
            node_id = item["node_id"]
            hostname = item["hostname"]
            is_core = item["is_master"]
            status = item["status"]

            click.secho("\n---------------------------------------")
            click.secho(f"Node Name: {name}")
            click.secho(f"{name} - Node ID: {node_id}")
            click.secho(f"{name} - Hostname: {hostname}")
            click.secho(f"{name} - IP: {ips}")
            click.secho(f"{name} - Is Core: {is_core}")
            click.secho(f"{name} - Status: {status}")
            click.secho(f"{name} - Last Seen: {last_seen}")
            click.secho(f"{name} - Is Environment Banner: {is_env}")
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""Test suite for assets."""
import pytest

from .base_assets import (AssetsPrivate, AssetsPublic, ModelMixinsBase,
                          check_assets, get_field_values, load_test_data)


class TestDevices(AssetsPrivate, AssetsPublic, ModelMixinsBase):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_devices):
        """Pass."""
        return load_test_data(apiobj=api_devices)

    def test_get_by_hostname(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_HOSTNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_hostname(
            value=value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)
        assert len(rows) == 1

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_hostname_equals_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_HOSTNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_hostname(
            value=value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_hostnames(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_HOSTNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_hostnames(
            values=values, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value in rows_values

    def test_get_by_hostnames_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_HOSTNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_values(
            values=values,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value not in rows_values

    def test_get_by_hostname_regex(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_HOSTNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_hostname_regex(
            value=regex_value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_hostname_regex_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_HOSTNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_hostname_regex(
            value=regex_value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_mac(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAC

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_mac(
            value=value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)
        assert len(rows) == 1

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_mac_equals_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAC

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_mac(
            value=value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_macs(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAC

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_macs(
            values=values, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value in rows_values

    def test_get_by_macs_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAC

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_values(
            values=values,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value not in rows_values

    def test_get_by_mac_regex(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAC

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_mac_regex(
            value=regex_value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_mac_regex_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAC

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_mac_regex(
            value=regex_value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_ip(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_IP

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_ip(
            value=value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)
        assert len(rows) >= 1

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_ip_equals_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_IP

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_ip(
            value=value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_ips(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_IP

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_ips(
            values=values, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value in rows_values

    def test_get_by_ips_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_IP

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_ips(
            values=values,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value not in rows_values

    def test_get_by_ip_regex(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_IP

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_ip_regex(
            value=regex_value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_ip_regex_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_IP

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_ip_regex(
            value=regex_value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_subnet(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_SUBNET

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_subnet(
            value=value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)
        assert len(rows) >= 1

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values



# -*- coding: utf-8 -*-
"""Test suite."""
import csv
import re
import sys
from io import StringIO

import pytest
from axonius_api_client import api, auth
from axonius_api_client.cli.context import Context
from axonius_api_client.http import Http
from click.testing import CliRunner

IS_WINDOWS = sys.platform == "win32"
IS_LINUX = sys.platform == "linux"
IS_MAC = sys.platform == "darwin"


def log_check(caplog, entries, exists=True):
    """Check if entries match caplog."""
    msgs = [rec.message for rec in caplog.records]
    for entry in entries:
        if exists:
            if not any(re.search(entry, m) for m in msgs):
                error = "Did not find entry in log: {!r}\nAll entries:\n{}"
                error = error.format(entry, "\n".join(msgs))
                raise Exception(error)
        else:
            if any(re.search(entry, m) for m in msgs):
                error = "Found unexpected entry in log: {!r}\nAll entries:\n{}"
                error = error.format(entry, "\n".join(msgs))
                raise Exception(error)


#
def get_cnx_existing(apiobj, name=None):
    """Pass."""
    found = None
    adapters = apiobj.get()
    for adapter in adapters:
        if name and adapter["name"] != name:
            continue
        cnxs = adapter["cnx"]
        for cnx in cnxs:
            found = cnxs[0]
            found["schemas"] = adapter["schemas"]["cnx"]
            break

    if not found:
        pytest.skip("No connections found for any adapter!")
    return found


def get_cnx_working(apiobj, name=None):
    """Pass."""
    problem_children = [
        "symantec_altiris",  # AX-7165
    ]

    found = None
    adapters = apiobj.get()
    for adapter in adapters:
        if name and adapter["name"] != name:
            continue
        if adapter["name"] in problem_children:
            continue
        cnxs = adapter["cnx"]
        for cnx in cnxs:
            if cnx["working"]:
                found = cnxs[0]
                found["schemas"] = adapter["schemas"]["cnx"]
                break

    if not found:
        pytest.skip("No working connections found for any adapter!")
    return found


def get_cnx_broken(apiobj, name=None):
    """Pass."""
    found = None
    adapters = apiobj.get()
    for adapter in adapters:
        if name and adapter["name"] != name:
            continue
        cnxs = adapter["cnx"]
        for cnx in cnxs:
            if not cnx["working"]:
                found = cnxs[0]
                found["schemas"] = adapter["schemas"]["cnx"]
                break

    if not found:
        pytest.skip("No broken connections found for any adapter!")
    return found


def get_url(request):
    """Pass."""
    return request.config.getoption("--ax-url").rstrip("/")


def get_key_creds(request):
    """Pass."""
    key = request.config.getoption("--ax-key")
    secret = request.config.getoption("--ax-secret")
    return {"key": key, "secret": secret}


def get_auth(request):
    """Pass."""
    http = Http(url=get_url(request), certwarn=False)

    obj = auth.ApiKey(http=http, **get_key_creds(request))
    obj.login()
    return obj


def check_apiobj(authobj, apiobj):
    """Pass."""
    url = authobj._http.url
    authclsname = format(authobj.__class__.__name__)
    assert authclsname in format(apiobj)
    assert authclsname in repr(apiobj)
    assert url in format(apiobj)
    assert url in repr(apiobj)

    assert isinstance(apiobj.auth, auth.Model)
    assert isinstance(apiobj.http, Http)
    assert isinstance(apiobj.router, api.routers.Router)


def check_apiobj_children(apiobj, **kwargs):
    """Pass."""
    for k, v in kwargs.items():
        attr = getattr(apiobj, k)
        attrclsname = format(attr.__class__.__name__)

        assert isinstance(attr, api.mixins.ChildMixins)
        assert isinstance(attr, v)

        assert isinstance(attr.auth, auth.Model)
        assert isinstance(attr.http, Http)
        assert isinstance(attr.router, api.routers.Router)
        assert isinstance(attr.parent, api.mixins.Model)
        assert attrclsname in format(attr)
        assert attrclsname in repr(attr)


def check_apiobj_xref(apiobj, **kwargs):
    """Pass."""
    for k, v in kwargs.items():
        attr = getattr(apiobj, k)

        assert isinstance(attr, api.mixins.ModelMixins)
        assert isinstance(attr, v)


def load_clirunner(request, monkeypatch):
    """Pass."""
    runner = CliRunner(mix_stderr=False)

    url = request.config.getoption("--ax-url")
    key = request.config.getoption("--ax-key")
    secret = request.config.getoption("--ax-secret")

    monkeypatch.setenv("AX_URL", url)
    monkeypatch.setenv("AX_KEY", key)
    monkeypatch.setenv("AX_SECRET", secret)
    return runner


def check_stderr_lines(result):
    """Pass."""
    stderr = result.stderr.splitlines()

    assert stderr[0].startswith("** WARNING: Unverified HTTPS request!")
    assert stderr[1].startswith("** Connected to "), stderr


class MockError(Exception):
    """Pass."""


class MockCtx:
    """Pass."""


def mock_failure(*args, **kwargs):
    """Pass."""
    raise MockError("badwolf")


def get_mockctx():
    """Pass."""
    ctx = MockCtx()
    ctx.obj = Context()
    return ctx


def check_csv_cols(content, cols):
    """Pass."""
    QUOTING = csv.QUOTE_NONNUMERIC
    fh = StringIO()
    fh.write(content)
    fh.seek(0)
    reader = csv.DictReader(fh, quoting=QUOTING)
    rows = []
    for row in reader:
        rows.append(row)
        for x in cols:
            assert x in row, "column {!r} not in {}".format(x, list(row))
    return rows



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options
from .grp_common import EXPORT, PERMS, ROLE_NAME, handle_export

OPTIONS = [*AUTH, EXPORT, ROLE_NAME, *PERMS]


@click.command(name="update", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, name, **kwargs):
    """Update a roles permissions."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = client.system.roles.update(name=name, **kwargs)
        ctx.obj.echo_ok(f"Updated role {name!r}")

    handle_export(ctx=ctx, data=data, export_format=export_format, **kwargs)



# -*- coding: utf-8 -*-
"""Python API Client for Axonius."""
from . import api, auth, cli, constants, exceptions, http, logs, tools, version
from .api import Adapters, Devices, Enforcements, System, Users
from .auth import ApiKey
from .connect import Connect
from .http import Http

__version__ = version.__version__
LOG = logs.LOG

__all__ = (
    # Connection handler
    "Connect",
    # http client
    "Http",
    # authentication
    "ApiKey",
    # api
    "Users",
    "Devices",
    "Adapters",
    "Enforcements",
    "System",
    # modules
    "api",
    "auth",
    "http",
    "exceptions",
    "version",
    "tools",
    "constants",
    "cli",
    "logs",
)



# -*- coding: utf-8 -*-
"""Test suite."""

import pytest
from axonius_api_client.exceptions import ApiError, NotFoundError

from ...meta import TEST_USER


@pytest.mark.skip(f"Waiting for update to 3.3!")  # XXX update public API for roles/users
class TestSystemUsers:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_system):
        """Pass."""
        return api_system.users

    def test__get(self, apiobj):
        """Pass."""
        data = apiobj._get()
        assert isinstance(data, list)
        for x in data:
            assert isinstance(x, dict)

    def test__get_limit(self, apiobj):
        """Pass."""
        data = apiobj._get(limit=1)
        assert isinstance(data, list)
        for x in data:
            assert isinstance(x, dict)
        assert len(data) == 1

    def test__get_limit_skip(self, apiobj):
        """Pass."""
        all_data = apiobj._get()
        data = apiobj._get(limit=1, skip=1)
        assert isinstance(data, list)
        for x in data:
            assert isinstance(x, dict)
        if len(all_data) == 1:
            assert len(data) == 0
        else:
            assert len(data) == 1

    def test_add_get_update_delete(self, apiobj):
        """Pass."""
        try:
            apiobj.get(name=TEST_USER)
        except NotFoundError:
            pass
        else:
            apiobj.delete(name=TEST_USER)

        added = apiobj.add(name=TEST_USER, password=TEST_USER)
        assert isinstance(added, dict)
        assert added["user_name"] == TEST_USER
        assert not added["first_name"]
        assert not added["last_name"]
        assert added["password"] == ["unchanged"]

        with pytest.raises(ApiError):
            apiobj.add(name=TEST_USER, password=TEST_USER)

        updated = apiobj.update(
            name=TEST_USER,
            firstname=TEST_USER,
            lastname=TEST_USER,
            password=TEST_USER + "X",
        )
        assert isinstance(updated, dict)
        assert updated["user_name"] == TEST_USER
        assert updated["first_name"] == TEST_USER
        assert updated["last_name"] == TEST_USER
        assert updated["password"] == ["unchanged"]

        roles = apiobj.parent.roles.get()
        rolename = roles[-1]["name"]
        updated_role = apiobj.update_role(name=TEST_USER, rolename=rolename)
        assert updated_role["role_name"] == rolename

        deleted = apiobj.delete(name=TEST_USER)
        assert isinstance(deleted, list)
        assert not [x for x in deleted if x["uuid"] == added["uuid"]]

        with pytest.raises(ApiError):
            apiobj.update(name=TEST_USER)

        with pytest.raises(NotFoundError):
            apiobj.get(name=TEST_USER)

    def test_add_bad_role(self, apiobj):
        """Pass."""
        val = "xxx"
        with pytest.raises(NotFoundError):
            apiobj.add(name=val, password=val, rolename="flimflam")



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options

METHOD = "delete-by-tags"
OPTIONS = [
    *AUTH,
    click.option(
        "--tag",
        "-t",
        "value",
        help="Tags of saved queries",
        required=True,
        multiple=True,
        show_envvar=True,
        show_default=True,
    ),
]


@click.command(name="delete-by-tags", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Delete saved queries by tags."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = apiobj.saved_query.get_by_tags(**kwargs)
        apiobj.saved_query.delete(rows=rows)

    ctx.obj.echo_ok("Successfully deleted saved queries:")

    for row in rows:
        ctx.obj.echo_ok("  {}".format(row["name"]))

    ctx.exit(0)



# -*- coding: utf-8 -*-
"""Utilities and tools."""
import json
import logging
import pathlib
import platform
import sys
from datetime import datetime, timedelta
from itertools import zip_longest
from urllib.parse import urljoin

import click
import dateutil.parser
import dateutil.relativedelta
import dateutil.tz

from . import __package__ as PACKAGE_ROOT
from .constants import (ERROR_ARGS, ERROR_TMPL, NO, OK_ARGS, OK_TMPL, SIMPLE,
                        WARN_ARGS, WARN_TMPL, YES)
from .exceptions import ApiError, ToolsError

LOG = logging.getLogger(PACKAGE_ROOT)

# XXX package time?
# def val_type(value, types):
#     """Check that value is one of types.

#     Notes:
#         * handy wrapper for isinstance to throw a wrapped exception

#     Args:
#         value (:obj:`object`): object to check type of
#         types (:obj:`tuple` of :class:`type`): tuple of types

#     Raises:
#         :exc:`ToolsError`: if value is not an instance of types

#     """
#     if not isinstance(value, types):
#         msg = "Invalid type for value {value!r}, must be one of {types!r}"
#         msg = msg.format(value=value, types=types)
#         raise ToolsError(msg)


def listify(obj, dictkeys=False):
    """Force an object into a list.

    Notes:
        * :obj:`list`: returns as is
        * :obj:`tuple`: convert to list
        * :obj:`None`: returns as an empty list
        * any of :data:`axonius_api_client.SIMPLE`: return as a list of obj
        * :obj:`dict`: if dictkeys is True, return as list of keys of obj,
          otherwise return as a list of obj

    Args:
        obj (:obj:`object`): object to coerce to list
        dictkeys (:obj:`bool`, optional): default ``False`` -

            * if ``True`` and obj is dict, return list of keys of obj
            * if ``False`` and obj is dict, return list of obj

    Returns:
        :obj:`list`: coerced list
    """
    if isinstance(obj, list):
        return obj

    if isinstance(obj, tuple):
        return list(obj)

    if obj is None:
        return []

    if isinstance(obj, SIMPLE):
        return [obj]

    if isinstance(obj, dict):
        if dictkeys:
            return list(obj)

        return [obj]

    return obj


def grouper(iterable, n, fillvalue=None):
    """Split an iterable into chunks.

    Args:
        iterable (:obj:`typing.Iterable`): iterable to split into chunks of size n
        n (:obj:`int`): length to split iterable into
        fillvalue (:obj:`object`, optional): default ``None`` - value to use
            as filler for last chunk

    Returns:
        :obj:`typing.Iterator`: an iterator with chunks of length n
    """
    return zip_longest(*([iter(iterable)] * n), fillvalue=fillvalue)


# def nest_depth(obj):
#     """Get the nesting depth of an object.

#     Notes:
#         * checks to see if how many complex sub-objects an object has.
#         * dictionary with any list or dict values would be 1.
#         * list with a dictionary where values are all simple would be 1.
#         * list of lists with simple values would be two.

#     Args:
#         obj (:obj:`object`): object to get the nesting depth of

#     Returns:
#         :obj:`int`: int of the complexity level of obj
#     """
#     if isinstance(obj, dict):
#         obj = list(obj.values())

#     if isinstance(obj, list):
#         calcs = [nest_depth(obj=x) for x in obj if isinstance(obj, COMPLEX)]
#         if calcs:
#             return 1 + max(calcs)
#         return 1
#     return 0


def coerce_int(obj):
    """Convert an object into int.

    Args:
        obj (:obj:`object`): object to convert to int

    Returns:
        :obj:`int`: coerced int

    Raises:
        :exc:`ToolsError`: if obj is not able to be converted to int
    """
    try:
        return int(obj)
    except Exception:
        raise ToolsError(f"Supplied value {obj!r} is not an integer.")


def coerce_bool(obj):
    """Convert an object into bool.

    Notes:
        * if obj is str, will check against :data:`axonius_api_client.YES`
          and :data:`axonius_api_client.NO`

    Args:
        obj (:obj:`object`): object to coerce to bool

    Returns:
        :obj:`bool`: coerced bool

    Raises:
        :exc:`ToolsError`: obj is not able to be converted to bool
    """
    coerce_obj = obj

    if isinstance(obj, str):
        coerce_obj = coerce_obj.lower().strip()

    if coerce_obj in YES:
        return True

    if coerce_obj in NO:
        return False

    msg = (
        f"Supplied value {coerce_obj!r} is not one of {YES} for true or {NO} for false."
    )
    raise ToolsError(msg)


def is_int(obj, digit=False):
    """Check if obj is int typeable.

    Args:
        obj (:obj:`object`): object to check
        digit (:obj:`bool`, optional): default ``False`` -
            * if ``True`` obj must be (str/bytes where isdigit is True) or int
            * if ``False`` obj must be int

    Returns:
        :obj:`bool`: bool reflecting if obj is int typeable
    """
    if digit:
        if (isinstance(obj, str) or isinstance(obj, bytes)) and obj.isdigit():
            return True

    return not isinstance(obj, bool) and isinstance(obj, int)


def join_url(url, *parts):
    """Join a URL to any number of parts.

    Args:
        url (:obj:`str`): str to add parts to
        *parts (:obj:`str`): str(s) to append to url

    Returns:
        :obj:`str`: url with parts appended
    """
    url = url.rstrip("/") + "/"
    for part in parts:
        if not part:
            continue
        url = url.rstrip("/") + "/"
        part = part.lstrip("/")
        url = urljoin(url, part)
    return url


# def join_dot(obj, empty=False, joiner="."):
#     """Join a string using periods.

#     Notes:
#         * obj will be coerced into a list
#         * if obj is dict, keys of dict will be used as obj

#     Args:
#         obj (:obj:`object` or :obj:`list` of :obj:`object`): objects(s) to convert
#             to str and join.
#         empty (:obj:`bool`, optional): default ``False`` -

#             * if ``True`` leave values in EMPTY in place
#             * if ``false`` remove values in EMPTY
#         joiner (:obj:`str`, optional): default ``"."`` - value to use when joining obj

#     Returns:
#         :obj:`str`: str of obj joined using joiner
#     """
#     obj = listify(obj=obj, dictkeys=True)

#     if not empty:
#         obj = [x for x in obj if x not in EMPTY and format(x)]

#     return joiner.join([format(x) for x in obj])


# def join_cr(obj, pre=True, post=False, indent="  ", joiner="\n"):
#     r"""Create str of elements joined by carriage return.

#     Notes:
#         * obj will be coerced into a list
#         * if obj is dict, keys of dict will be used as obj

#     Args:
#         obj (:obj:`object` or :obj:`list` of :obj:`object`): objects(s) to convert to
#             str and join
#         pre (:obj:`bool`, optional): default ``True`` -

#             * if ``True`` add joiner to the beginning of the joined str
#             * if ``False`` do not add joiner to the beginning of the joined str

#         post (:obj:`bool`, optional): default ``False`` -

#             * if ``True`` add joiner to the end of the joined str
#             * if ``False`` do not add joiner to the end of the joined str
#         indent (:obj:`str`, optional): default ``"  "`` - value to prefix joiner
#         joiner (:obj:`str`, optional): default ``"\n"`` - value to use when joining obj

#     Returns:
#         :obj:`str`: str of obj joined using joiner
#     """
#     obj = listify(obj=obj, dictkeys=True)

#     if indent:
#         joiner = "{}{}".format(joiner, indent)

#     joined = joiner.join([format(x) for x in obj])

#     if joined:
#         if pre:
#             joined = joiner + joined
#         if post:
#             joined = joined + joiner

#     return joined


# def join_comma(obj, empty=False, indent=" ", joiner=","):
#     """Create str of elements joined by comma.

#     Notes:
#         * obj will be coerced into a list
#         * if obj is dict, keys of dict will be used as obj

#     Args:
#         obj (:obj:`object` or :obj:`list` of :obj:`object`): objects(s) to convert
#             to str and join.
#         empty (:obj:`bool`, optional): default ``False`` -

#             * if ``True`` leave values in EMPTY in place
#             * if ``False`` remove values in EMPTY
#         indent (:obj:`str`, optional): default ``"  "`` - value to prefix joiner
#         joiner (:obj:`str`, optional): default ``","`` - value to use when joining obj

#     Returns:
#         :obj:`str`: str of obj joined using joiner
#     """
#     obj = listify(obj=obj, dictkeys=True)

#     if not empty:
#         obj = [x for x in obj if x not in EMPTY and format(x)]

#     if indent:
#         joiner = "{}{}".format(joiner, indent)

#     return joiner.join([format(x) for x in obj])


def strip_right(obj, fix):
    """Strip text from the right side of obj.

    Args:
        obj (:obj:`str`) or (:obj:`list` of :obj:`str`): str(s) to strip fix from
        fix (:obj:`str`): str to remove from obj(s)

    Returns:
        (:obj:`str`) or (:obj:`list` of :obj:`str`): obj with fix stripped
    """
    if isinstance(obj, list) and all([isinstance(x, str) for x in obj]):
        return [strip_right(obj=x, fix=fix) for x in obj]

    if isinstance(obj, str):
        plen = len(fix)

        if obj.endswith(fix):
            return obj[:-plen]

    return obj


def strip_left(obj, fix):
    """Strip text from the left side of obj.

    Args:
        obj (:obj:`str`) or (:obj:`list` of :obj:`str`): str(s) to strip fix from
        fix (:obj:`str`): str to remove from obj(s)

    Returns:
        (:obj:`str`) or (:obj:`list` of :obj:`str`): obj with fix stripped
    """
    if isinstance(obj, list) and all([isinstance(x, str) for x in obj]):
        return [strip_left(obj=x, fix=fix) for x in obj]

    if isinstance(obj, str):
        plen = len(fix)

        if obj.startswith(fix):
            return obj[plen:]

    return obj


def json_dump(obj, indent=2, sort_keys=False, error=True, **kwargs):
    """Serialize an object into json str.

    Args:
        obj (:obj:`object`): object to serialize into json str
        indent (:obj:`int`, optional): default ``2`` - json str indentation
        sort_keys (:obj:`bool`, optional): default ``False`` -

            * if ``True`` sort keys of dicts
            * if ``False`` leave keys of dicts sorted as is
        error (:obj:`bool`, optional): default ``True`` -

            * if ``True`` If json error happens, raise it
            * if ``False`` If json error happens, return original obj as is
        **kwargs: passed to :func:`json.dumps`

    Returns:
        (:obj:`str`) or (:obj:`object`):
            * object: if json error happens and error is false
            * str: if no json error happens
    """
    if isinstance(obj, bytes):
        obj = obj.decode("utf-8")

    try:
        return json.dumps(obj, indent=indent, sort_keys=sort_keys, **kwargs)
    except Exception:
        if error:
            raise
        return obj


def json_load(obj, error=True, **kwargs):
    """Deserialize a json str into an object.

    Args:
        obj (:obj:`object`): str to deserialize into obj
        error (:obj:`bool`, optional): default ``True`` -

            * if ``True`` If json error happens, raise it
            * if ``False`` If json error happens, return original obj as is
        **kwargs: passed to :func:`json.loads`

    Returns:
        :obj:`object`: json str serialized into python obj
    """
    try:
        return json.loads(obj, **kwargs)
    except Exception:
        if error:
            raise
        return obj


def json_reload(obj, error=False, trim=None, **kwargs):
    """Re-serialize a json str into a pretty json str.

    Args:
        obj (:obj:`object`): str to deserialize into obj and serialize back to str
        error (:obj:`bool`, optional): default ``False`` -

            * if ``True`` If json error happens, raise it
            * if ``False`` If json error happens, return original obj as is
        **kwargs: passed to :func:`json_dump`

    Returns:
        (:obj:`str`) or (:obj:`object`):
            * object: if json error happens and error is false
            * str: if no json error happens
    """
    obj = json_load(obj=obj, error=error)
    if not isinstance(obj, str):
        obj = json_dump(obj=obj, error=error, **kwargs)
    obj = obj or ""
    if isinstance(obj, str):
        obj = obj.strip()
        if trim and len(obj) >= trim:
            obj = obj[:trim] + f"\nTrimmed over {trim} characters"
    return obj


def dt_parse(obj):
    """Parse a str, datetime, or timedelta into a datetime object.

    Notes:
        * :obj:`str`: will be parsed into datetime obj
        * :obj:`datetime.timedelta`: will be parsed into datetime obj as now - timedelta
        * :obj:`datetime.datetime`: will be re-parsed into datetime obj

    Args:
        obj (:obj:`str` or :obj:`datetime.datetime` or :obj:`datetime.timedelta`):
            object or list of objects to parse

    Returns:
        :obj:`datetime.datetime`: parsed datetime
    """
    if isinstance(obj, list) and all([isinstance(x, str) for x in obj]):
        return [dt_parse(obj=x) for x in obj]

    if isinstance(obj, datetime):
        obj = str(obj)

    if isinstance(obj, timedelta):
        obj = str(dt_now() - obj)

    return dateutil.parser.parse(obj)


def dt_now(delta=None, tz=dateutil.tz.tzutc()):
    """Get the current datetime in for a specific tz.

    Args:
        delta (:obj:`datetime.timedelta`, optional): default ``None`` -
            subtract timedelta from now
        tz (:obj:`datetime.timezone`, optional): default :obj:`dateutil.tz.tzutc`:
            get now for a specific timezone

    Returns:
        :obj:`datetime.datetime`: datetime of now
    """
    if isinstance(delta, timedelta):
        return dt_parse(obj=delta)
    return datetime.now(tz)


def dt_sec_ago(obj):
    """Get number of seconds ago a given datetime was.

    Args:
        obj (:obj:`object`): will be parsed by :meth:`dt_parse` into a datetime obj

    Returns:
        :obj:`int`:
            int of secs ago obj was
    """
    obj = dt_parse(obj=obj)
    now = dt_now(tz=obj.tzinfo)
    return round((now - obj).total_seconds())


def dt_min_ago(obj):
    """Get number of minutes ago a given datetime was.

    Args:
        obj (:obj:`object`): will be parsed by :meth:`dt_sec_ago` into seconds ago

    Returns:
        :obj:`int`: int of minutes ago obj was
    """
    return round(dt_sec_ago(obj=obj) / 60)


def dt_within_min(obj, n=None):
    """Check if given datetime is within the past n minutes.

    Args:
        obj (:obj:`object`): will be parsed by :meth:`dt_min_ago` into minutes ago
        n (:obj:`int` or :obj:`str`, optional): default ``None`` - int of
            dt_min_ago(obj) should be greater than or equal to

    Returns:
        :obj:`bool`: bool representing if obj is within the past n minutes
    """
    if not is_int(obj=n, digit=True):
        return False

    return dt_min_ago(obj=obj) >= int(n)


def get_path(obj):
    """Convert a str into a fully resolved & expanded Path object.

    Args:
        obj (:obj:`str` or :obj:`pathlib.Path`): obj to convert into
            expanded and resolved absolute Path obj

    Returns:
        :obj:`pathlib.Path`: resolved path
    """
    return pathlib.Path(obj).expanduser().resolve()


def path_read(obj, binary=False, is_json=False, **kwargs):
    """Read data from a file.

    Notes:
        * obj will be parsed by :meth:`path`
        * if path filename ends with ".json", data will be deserialized using
          :meth:`json_load`

    Args:
        obj (:obj:`str` or :obj:`pathlib.Path`): path to read data form
        binary (:obj:`bool`, optional): default ``False`` -

            * if ``True`` read the data as binary
            * if ``False`` read the data as str
        is_json (:obj:`bool`, optional): default ``False`` -
            * if ``True`` deserialize data using :meth:`json_load`
            * if ``False`` do not deserialize data
        **kwargs: passed to :meth:`json_load`

    Returns:
        :obj:`tuple` of (:obj:`pathlib.Path`, :obj:`object`):
            resolved path and data read from path

    Raises:
        :exc:`ToolsError`: path does not exist as file
    """
    robj = get_path(obj=obj)

    if not robj.is_file():
        raise ToolsError(f"Supplied path='{obj}' (resolved='{robj}') does not exist!")

    if binary:
        data = robj.read_bytes()
    else:
        data = robj.read_text()

    if is_json:
        data = json_load(obj=data, **kwargs)

    if robj.suffix == ".json" and isinstance(data, str):
        kwargs.setdefault("error", False)
        data = json_load(obj=data, **kwargs)

    return robj, data


def path_write(
    obj,
    data,
    overwrite=False,
    binary=False,
    binary_encoding="utf-8",
    is_json=False,
    make_parent=True,
    protect_file=0o600,
    protect_parent=0o700,
    # fmt: off
    **kwargs
    # fmt: on
):
    """Write data to a file.

    Notes:
        * obj will be parsed by :meth:`path`
        * if obj filename ends with ".json", serializes data using :meth:`json_dump`.

    Args:
        obj (:obj:`str` or :obj:`pathlib.Path`): path to write data to
        data (:obj:`str` or :obj:`bytes` or :obj:`object`): data to write to obj
        overwrite (:obj:`bool`, optional): default ``False`` -

            * if ``True`` overwrite obj
            * if ``False`` do not overwrite obj, raise exception if it already exists
        binary (:obj:`bool`, optional): default ``False`` -

            * if ``True`` write the data as binary
            * if ``False`` write the data as str
        binary_encoding (:obj:`str`, optional): default ``"utf-8"`` - encoding to
            use when switching from str/bytes
        is_json (:obj:`bool`, optional): default ``False`` -

            * if ``True`` Serialize data using :meth:`json_load` before writing
            * if ``False`` Do not serialize data before writing
        make_parent (:obj:`bool`, optional): default ``True`` -

            * if ``True`` If the parent directory does not exist, create it
            * if ``False`` If the parent directory does not exist, raise exception
        protect_file (:obj:`oct`): default ``0o600`` (read/write to owner only) -
            octal mode of permissions to set on file
        protect_dir (:obj:`oct`): default ``0o700`` (read/write/execute to owner only) -
            octal mode of permissions to set on parent directory when creating
        **kwargs: passed to :meth:`json_dump`

    Returns:
        :obj:`tuple` of (:obj:`pathlib.Path`, method):
            tuple of (resolved path obj, method used to write data to file)

    Raises:
        :exc:`ToolsError`: path exists as file and overwrite is False
        :exc:`ToolsError`: if parent path does not exist and make_parent
            is False
    """
    obj = get_path(obj=obj)

    if is_json:
        data = json_dump(obj=data, **kwargs)

    if obj.suffix == ".json" and not (isinstance(data, str) or isinstance(data, bytes)):
        kwargs.setdefault("error", False)
        data = json_dump(obj=data, **kwargs)

    if binary:
        if isinstance(data, str):
            data = data.encode(binary_encoding)
        method = obj.write_bytes
    else:
        if isinstance(data, bytes):
            data = data.decode(binary_encoding)
        method = obj.write_text

    if obj.is_file() and overwrite is False:
        raise ToolsError(f"File '{obj}' already exists and overwrite is False")

    if not obj.parent.is_dir():
        if make_parent:
            obj.parent.mkdir(mode=protect_parent, parents=True, exist_ok=True)
        else:
            error = f"Directory '{obj.parent}' does not exist and make_parent is False"
            raise ToolsError(error)

    obj.touch()

    if protect_file:
        obj.chmod(protect_file)

    return obj, method(data)


def longest_str(obj):
    """Badwolf."""
    return round(max([len(x) + 5 for x in obj]), -1)


def split_str(obj, split=",", strip=None, do_strip=True, lower=True, empty=False):
    """Split a string or list of strings into a list of strings."""
    if isinstance(obj, list):
        return [
            y
            for x in obj
            for y in split_str(
                obj=x,
                split=split,
                strip=strip,
                do_strip=do_strip,
                lower=lower,
                empty=empty,
            )
        ]
    if not isinstance(obj, str):
        raise ApiError(f"Unable to split non-str value {obj}")

    ret = []
    for x in obj.split(split):
        if lower:
            x = x.lower()
        if do_strip:
            x = x.strip(strip)
        if not empty and not x:
            continue
        ret.append(x)
    return ret


def echo_ok(msg, tmpl=True, **kwargs):
    """Pass."""
    echoargs = {}
    echoargs.update(OK_ARGS)
    echoargs.update(kwargs)
    if tmpl:
        msg = OK_TMPL.format(msg=msg)

    LOG.info(msg)
    click.secho(msg, **echoargs)


def echo_warn(msg, tmpl=True, **kwargs):
    """Pass."""
    echoargs = {}
    echoargs.update(WARN_ARGS)
    echoargs.update(kwargs)
    if tmpl:
        msg = WARN_TMPL.format(msg=msg)

    LOG.warning(msg)
    click.secho(msg, **echoargs)


def echo_error(msg, abort=True, tmpl=True, **kwargs):
    """Pass."""
    echoargs = {}
    echoargs.update(ERROR_ARGS)
    echoargs.update(kwargs)
    if tmpl:
        msg = ERROR_TMPL.format(msg=msg)

    LOG.error(msg)
    click.secho(msg, **echoargs)
    if abort:
        sys.exit(1)


def sysinfo():
    """Pass."""
    info = {}
    info["Date"] = str(dt_now())
    info["Python System Version"] = ", ".join(sys.version.splitlines())
    platform_attrs = [
        "machine",
        "node",
        "platform",
        "processor",
        "python_branch",
        "python_compiler",
        "python_implementation",
        "python_revision",
        "python_version",
        "release",
        "system",
        "version",
        "win32_edition",
    ]
    for attr in platform_attrs:
        method = getattr(platform, attr)
        value = method()
        attr = attr.replace("_", " ").title()
        info[attr] = value
    return info



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
import logging

import pytest

from axonius_api_client.connect import Connect
from axonius_api_client.exceptions import ConnectError, InvalidCredentials
from axonius_api_client.http import requests

from ..utils import IS_LINUX, get_key_creds, get_url

BAD_CRED = "tardis"


class TestConnect:
    """Pass."""

    def test_no_start(self, request):
        """Pass."""
        ax_url = get_url(request)

        c = Connect(url=ax_url, key=BAD_CRED, secret=BAD_CRED)

        assert "Not connected" in format(c)
        assert "Not connected" in repr(c)
        assert c._handler_file is None
        assert c._handler_con is None

    def test_no_start_logs(self, request):
        """Pass."""
        ax_url = get_url(request)

        c = Connect(
            url=ax_url, key=BAD_CRED, secret=BAD_CRED, log_console=True, log_file=True
        )

        assert "Not connected" in format(c)
        assert "Not connected" in repr(c)
        assert isinstance(c._handler_file, logging.Handler)
        assert isinstance(c._handler_con, logging.Handler)

    def test_start(self, request):
        """Pass."""
        ax_url = get_url(request)

        c = Connect(url=ax_url, certwarn=False, **get_key_creds(request))

        c.start()

        assert "Connected" in format(c)
        assert "Connected" in repr(c)

        format(c.system.settings_lifecycle)
        format(c.system.settings_gui)
        format(c.system.settings_core)
        format(c.system.users)
        format(c.system.roles)
        format(c.system.nodes)
        format(c.system.discover)
        format(c.system.meta)
        format(c.enforcements)
        format(c.users)
        format(c.devices)
        format(c.adapters)

    def test_invalid_creds(self, request):
        """Pass."""
        ax_url = get_url(request)

        c = Connect(url=ax_url, key=BAD_CRED, secret=BAD_CRED, certwarn=False)

        c._http.CONNECT_TIMEOUT = 1

        with pytest.raises(ConnectError) as exc:
            c.start()

        assert isinstance(exc.value.exc, InvalidCredentials)

    def test_connect_timeout(self):
        """Pass."""
        c = Connect(url="127.0.0.99", key=BAD_CRED, secret=BAD_CRED, certwarn=False)

        c._http.CONNECT_TIMEOUT = 1

        with pytest.raises(ConnectError) as exc:
            c.start()

        if IS_LINUX:
            assert isinstance(exc.value.exc, requests.ConnectionError)
        else:
            assert isinstance(exc.value.exc, requests.ConnectTimeout)

    def test_connect_error(self):
        """Pass."""
        c = Connect(
            url="https://127.0.0.1:3919", key=BAD_CRED, secret=BAD_CRED, certwarn=False
        )

        c._http.CONNECT_TIMEOUT = 1

        with pytest.raises(ConnectError) as exc:
            c.start()
        assert isinstance(exc.value.exc, requests.ConnectionError)

    def test_invalid_creds_nowrap(self, request):
        """Pass."""
        ax_url = get_url(request)

        c = Connect(
            url=ax_url, key=BAD_CRED, secret=BAD_CRED, certwarn=False, wraperror=False
        )

        c._http.CONNECT_TIMEOUT = 1

        with pytest.raises(InvalidCredentials):
            c.start()

    def test_other_exc(self, request):
        """Pass."""
        c = Connect(url="127.0.0.1", key=BAD_CRED, secret=BAD_CRED, certwarn=False)

        c._http.CONNECT_TIMEOUT = 1
        c._auth._creds = None

        with pytest.raises(ConnectError):
            c.start()

    def test_reason(self):
        """Pass."""
        exc = Exception("badwolf")

        reason = Connect._get_exc_reason(exc)

        assert format(reason) == "badwolf"



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import click

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json-full", "json-config", "str"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)

SECTION = click.option(
    "--section",
    "-se",
    "section",
    help="Settings section internal name (not title)",
    required=True,
    show_envvar=True,
    show_default=True,
)

SUB_SECTION = click.option(
    "--sub-section",
    "-sb",
    "sub_section",
    help="Settings sub section internal name (not title)",
    required=True,
    show_envvar=True,
    show_default=True,
)


def str_sections(meta):
    """Pass."""
    settings_title = meta["settings_title"]
    click.secho(f"Settings: {settings_title}")

    sections = meta["sections"]

    for name, meta in sections.items():
        str_section(meta=meta)


def str_section(meta):
    """Pass."""
    name = meta["name"]
    title = meta["title"]
    settings_title = meta["settings_title"]

    lines = [
        f"Section Name: {name!r}",
        f"Section Title: {settings_title}: {title}",
    ]
    join = "\n- "
    click.secho(join + join.join(lines))
    str_schemas(schemas=meta["schemas"], config=meta["config"], indent=4)

    sub_sections = meta["sub_sections"]
    for sub_name, sub_meta in sub_sections.items():
        str_subsection(meta=sub_meta)


def str_subsection(meta):
    """Pass."""
    name = meta["name"]
    title = meta["title"]
    settings_title = meta["settings_title"]
    parent_title = meta["parent_title"]

    lines = [
        f"Sub Section Name: {name!r}",
        f"Sub Section Title: {settings_title}: {parent_title}: {title}",
    ]
    join = "\n--- "
    click.secho(join + join.join(lines))
    str_schemas(schemas=meta["schemas"], config=meta["config"], indent=6)


def str_schemas(schemas, config, indent=4):
    """Pass."""
    join = "\n{}".format(" " * indent)
    for name, schema in schemas.items():
        title = schema.get("title")
        stype = schema["type"]
        value = config[name]

        lines = [
            f"Name: {name}",
            f"Title: {title}",
            f"Type: {stype!r}",
            f"Value: {name}={value!r}",
        ]
        click.secho(join + join.join(lines))



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, NODE_CNX, SPLIT_CONFIG_OPT, add_options
from .grp_common import PROMPTS, prompt_config

OPTIONS = [
    *AUTH,
    *NODE_CNX,
    *PROMPTS,
    SPLIT_CONFIG_OPT,
]


@click.command(name="test", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, config, adapter_name, adapter_node, **kwargs):
    """Test reachability for an adapter from prompts or arguments."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    new_config = dict(config)
    prompt_config(
        ctx=ctx,
        client=client,
        new_config=new_config,
        adapter_name=adapter_name,
        adapter_node=adapter_node,
        **kwargs,
    )

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        client.adapters.cnx.test(
            adapter_name=adapter_name, adapter_node=adapter_node, **new_config,
        )
        ctx.obj.echo_ok(msg=f"Reachability test succeeded!")



# -*- coding: utf-8 -*-
"""API model for working with system configuration."""
import math

from ..mixins import ChildMixins


class Meta(ChildMixins):
    """Child API model for working with instance metadata."""

    def about(self):
        """Get about page metadata.

        Returns:
            :obj:`dict`: about page metadata
        """
        return self._about()

    def historical_sizes(self):
        """Get disk usage metadata.

        Returns:
            :obj:`dict`: disk usage metadata
        """
        return parse_sizes(self._historical_sizes())

    @property
    def version(self):
        """Get the version of Axonius."""
        about = self.about()
        return (about["Version"] or "DEMO").replace("_", ".")

    def _init(self, parent):
        """Post init method for subclasses to use for extra setup.

        Args:
            parent (:obj:`.api.mixins.Model`): parent API model of this child
        """
        super(Meta, self)._init(parent=parent)

    def _about(self):
        """Direct API method to get the About page.

        Returns:
            :obj:`dict`: about page metadata
        """
        path = self.router.meta_about
        return self.request(method="get", path=path)

    def _historical_sizes(self):
        """Direct API method to get the metadata about disk usage.

        Returns:
            :obj:`dict`: disk usage metadata
        """
        path = self.router.meta_historical_sizes
        return self.request(method="get", path=path)


def parse_sizes(raw):
    """Pass."""
    parsed = {}
    parsed["disk_free_mb"] = math.floor(raw["disk_free"] / 1024 / 1024)
    parsed["disk_used_mb"] = math.ceil(raw["disk_used"] / 1024 / 1024)
    parsed["historical_sizes_devices"] = raw["entity_sizes"].get("Devices", {})
    parsed["historical_sizes_users"] = raw["entity_sizes"].get("Users", {})
    return parsed



# -*- coding: utf-8 -*-
"""Test suite for assets."""

import copy
import io

import pytest
from axonius_api_client.constants import AGG_ADAPTER_NAME
from axonius_api_client.exceptions import ApiError

from .callbacks import Callbacks, load_test_data


class CallbacksTable(Callbacks):
    """Pass."""

    @pytest.fixture(scope="class")
    def cbexport(self):
        """Pass."""
        return "table"

    def test_row_as_is(self, cbexport, apiobj):
        """Pass."""
        field_complex = apiobj.TEST_DATA["field_complex"]

        if field_complex is None:
            pytest.skip(f"No complex field found for {apiobj}")

        schema = apiobj.fields.get_field_schema(
            value=field_complex,
            schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME],
        )
        sub_titles = [x["column_title"] for x in schema["sub_fields"] if x["is_root"]]

        io_fd = io.StringIO()

        getargs = {"export_fd": io_fd}

        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.start()

        for row in copy.deepcopy(apiobj.TEST_DATA["cb_assets"][:200]):
            rows_ret = cbobj.row(row=copy.deepcopy(row))
            assert len(rows_ret) == 1
            assert "Aggregated: Asset Unique ID" in rows_ret[0]
            assert schema["name_qual"] not in rows_ret[0]
            for i in sub_titles:
                assert i in rows_ret[0]

        cbobj.stop()
        output = io_fd.getvalue()
        checklines = "\n".join(output.splitlines()[:3])
        for i in sub_titles:
            assert i in checklines

    def test_check_fmt(self, cbexport, apiobj):
        """Pass."""
        getargs = {}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        with pytest.raises(ApiError):
            cbobj.check_fmt("badwolf")

    def test_check_stop(self, cbexport, apiobj):
        """Pass."""
        getargs = {"table_max_rows": 10}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.STATE["rows_processed"] = 10
        cbobj.check_stop()
        assert cbobj.STATE["stop"]
        assert cbobj.STATE["stop_msg"]


class TestDevicesCallbacksTable(CallbacksTable):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_devices):
        """Pass."""
        return load_test_data(apiobj=api_devices)


class TestUsersCallbacksTable(CallbacksTable):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_users):
        """Pass."""
        return load_test_data(apiobj=api_users)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, add_options

OPTIONS = [*AUTH]


@click.command(name="get-tags", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Get all known tags."""
    p_grp = ctx.parent.command.name

    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    apiobj = getattr(client, p_grp)
    apimethod = apiobj.labels.get

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = apimethod()

    for row in rows:
        click.secho(row)



# -*- coding: utf-8 -*-
"""Test suite for assets."""
import pytest

from .base_assets import (
    AssetsPrivate,
    AssetsPublic,
    ModelMixinsBase,
    check_assets,
    get_field_values,
    load_test_data,
)


class TestUsers(AssetsPrivate, AssetsPublic, ModelMixinsBase):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_users):
        """Pass."""
        return load_test_data(apiobj=api_users)

    def test_get_by_username(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_USERNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_username(
            value=value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)
        assert len(rows) == 1

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_username_equals_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_USERNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_username(
            value=value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_usernames(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_USERNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_usernames(
            values=values, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value in rows_values

    def test_get_by_usernames_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_USERNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_usernames(
            values=values,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value not in rows_values

    def test_get_by_username_regex(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_USERNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_username_regex(
            value=regex_value, fields_map=apiobj.TEST_DATA["fields_map"], field=field,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_username_regex_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_USERNAME

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_username_regex(
            value=regex_value,
            field=field,
            not_flag=True,
            fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_mail(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAIL

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_mail(
            value=value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)
        assert len(rows) >= 1

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_mail_equals_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAIL

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]

        rows = apiobj.get_by_mail(
            value=value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values

    def test_get_by_mails(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAIL

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_mails(
            values=values, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value in rows_values

    def test_get_by_mails_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAIL

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        values = values[0:2]

        rows = apiobj.get_by_mails(
            values=values,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        for value in values:
            assert value not in rows_values

    def test_get_by_mail_regex(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAIL

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_mail_regex(
            value=regex_value, field=field, fields_map=apiobj.TEST_DATA["fields_map"],
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value in rows_values

    def test_get_by_mail_regex_not(self, apiobj):
        """Pass."""
        field = apiobj.FIELD_MAIL

        values = get_field_values(rows=apiobj.TEST_DATA["assets"], field=field)
        value = values[0]
        regex_value = value[0:5]

        rows = apiobj.get_by_mail_regex(
            value=regex_value,
            field=field,
            fields_map=apiobj.TEST_DATA["fields_map"],
            not_flag=True,
        )
        check_assets(rows)

        rows_values = get_field_values(rows=rows, field=field)
        assert value not in rows_values



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import json_dump
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, INPUT_FILE, add_options
from .grp_common import EXPORT, SECTION, SUB_SECTION, str_section

OPTIONS = [*AUTH, EXPORT, SECTION, SUB_SECTION, INPUT_FILE]


@click.command(name="update-sub-section-from-json", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(
    ctx, url, key, secret, input_file, section, sub_section, export_format, **kwargs,
):
    """Update a sub-section from a JSON file."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    new_config = ctx.obj.read_stream_json(stream=input_file, expect=dict)

    apiname = ctx.parent.command.name.replace("-", "_")
    apiobj = getattr(client.system, apiname)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        settings = apiobj.update_sub_section(
            section=section, sub_section=sub_section, **new_config
        )
        ctx.obj.echo_ok(f"Updated {sub_section!r} with configuration {new_config}")

    if export_format == "str":
        str_section(meta=settings)
        ctx.exit(0)

    if export_format == "json-config":
        config = settings["config"]
        click.secho(json_dump(config))
        ctx.exit(0)

    if export_format == "json-full":
        click.secho(json_dump(settings))
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.http."""
import logging
import sys

import pytest
import requests

from axonius_api_client.exceptions import HttpError
from axonius_api_client.http import Http, ParserUrl
from axonius_api_client.version import __version__

from ..meta import (
    TEST_CLIENT_CERT,
    TEST_CLIENT_CERT_NAME,
    TEST_CLIENT_KEY,
    TEST_CLIENT_KEY_NAME,
)
from ..utils import get_url, log_check

InsecureRequestWarning = requests.urllib3.exceptions.InsecureRequestWarning


class TestParserUrl:
    """Test ParserUrl."""

    def test_schemehostport443(self):
        """Test a proper URL gets parsed the same."""
        u = ParserUrl("https://host:443/blah")
        assert u.hostname == "host"
        assert u.port == 443
        assert u.scheme == "https"
        assert u.parsed.path == "/blah"
        assert u.url_full == "https://host:443/blah"
        assert u.url == "https://host:443"

    def test_str_repr(self):
        """Test str/repr has URL path."""
        u = ParserUrl("https://host:443/blah")
        assert u.parsed.path in format(u)
        assert u.parsed.path in repr(u)

    def test_schemehost_noport443(self):
        """Test port gets added for https scheme."""
        u = ParserUrl("https://host")
        assert u.hostname == "host"
        assert u.port == 443
        assert u.scheme == "https"

    def test_justhost(self):
        """Test schema added for just host."""
        u = ParserUrl("host")
        assert u.hostname == "host"
        assert u.port == 443
        assert u.scheme == "https"

    def test_justhostport(self):
        """Test schema added for just host and port."""
        u = ParserUrl("host:443")
        assert u.hostname == "host"
        assert u.port == 443
        assert u.scheme == "https"

    def test_host_noschemeport(self):
        """Test exc when no port or scheme in URL."""
        exc = HttpError
        match = "no.*'port'"
        with pytest.raises(exc, match=match):
            ParserUrl("host", default_scheme="")

    def test_unknownschemehost_noport(self):
        """Test exc when no port and non http/https scheme."""
        exc = HttpError
        match = "no.*'port'"
        with pytest.raises(exc, match=match):
            ParserUrl("httpx://host")

    def test_hostport443_withslash(self):
        """Test scheme added with port 443 and no scheme in URL."""
        u = ParserUrl("host:443/")
        assert u.hostname == "host"
        assert u.port == 443
        assert u.scheme == "https"

    def test_hostport443_noscheme(self):
        """Test scheme added with port 443 and no scheme in URL."""
        u = ParserUrl("host:443", default_scheme="")
        assert u.hostname == "host"
        assert u.port == 443
        assert u.scheme == "https"

    def test_hostport80_noscheme(self):
        """Test scheme added with port 80 and no scheme in URL."""
        u = ParserUrl("host:80", default_scheme="")
        assert u.hostname == "host"
        assert u.port == 80
        assert u.scheme == "http"

    def test_schemehost_noport80(self):
        """Test port added with no port and http scheme in URL."""
        u = ParserUrl("http://host")
        assert u.hostname == "host"
        assert u.port == 80
        assert u.scheme == "http"


class TestHttp:
    """Test Http."""

    def test_str_repr(self, request):
        """Test str/repr has URL."""
        ax_url = get_url(request)

        http = Http(url=ax_url)

        assert ax_url in format(http)
        assert ax_url in repr(http)

    def test_parsed_url(self, request):
        """Test url=ParserUrl() works."""
        ax_url = get_url(request)

        parsed_url = ParserUrl(url=ax_url, default_scheme="https")

        http = Http(url=parsed_url)

        assert ax_url in format(http)
        assert ax_url in repr(http)

    def test_user_agent(self, request):
        """Test user_agent has version in it."""
        ax_url = get_url(request)

        http = Http(url=ax_url)
        assert __version__ in http.user_agent

    def test_certwarn_true(self, request, httpbin_secure):
        """Test quiet_urllib=False shows warning from urllib3."""
        url = httpbin_secure.url
        http = Http(url=url, certwarn=True, save_history=True)

        with pytest.warns(InsecureRequestWarning):
            http()

    def test_certwarn_false(self, request, httpbin_secure):
        """Test quiet_urllib=False shows warning from urllib3."""
        url = httpbin_secure.url
        http = Http(url=url, certwarn=False)

        http()

    @pytest.mark.skipif(sys.version_info < (3, 6), reason="requires python3.6 or higher")
    def test_verify_ca_bundle(self, request, httpbin_secure, httpbin_ca_bundle):
        """Test quiet_urllib=False no warning from urllib3 when using ca bundle."""
        url = httpbin_secure.url
        http = Http(url=url, certwarn=False)
        response = http()
        assert response.status_code == 200

    def test_save_last_true(self, request):
        """Test last req/resp with save_last=True."""
        ax_url = get_url(request)

        http = Http(url=ax_url, save_last=True, certwarn=False)
        response = http()
        assert response == http.LAST_RESPONSE
        assert response.request == http.LAST_REQUEST

    def test_save_last_false(self, request):
        """Test last req/resp with save_last=False."""
        ax_url = get_url(request)

        http = Http(url=ax_url, save_last=False, certwarn=False)

        http()

        assert not http.LAST_RESPONSE
        assert not http.LAST_REQUEST

    def test_save_history(self, request):
        """Test last resp added to history with save_history=True."""
        ax_url = get_url(request)

        http = Http(url=ax_url, save_history=True, certwarn=False)

        response = http()

        assert response in http.HISTORY

    def test_client_cert_missing_one(self, request, tmp_path):
        """Test cert or key supplied, but not the other."""
        ax_url = get_url(request)
        test_path = tmp_path / TEST_CLIENT_CERT_NAME
        test_path.write_text(TEST_CLIENT_CERT)

        with pytest.raises(HttpError):
            Http(url=ax_url, cert_client_cert=test_path, certwarn=False)

        with pytest.raises(HttpError):
            Http(url=ax_url, cert_client_key=test_path, certwarn=False)

    def test_client_cert_seperate(self, request, tmp_path):
        """Test cert or key supplied, but not the other."""
        ax_url = get_url(request)
        cert_path = tmp_path / TEST_CLIENT_CERT_NAME
        cert_path.write_text(TEST_CLIENT_CERT)
        key_path = tmp_path / TEST_CLIENT_KEY_NAME
        key_path.write_text(TEST_CLIENT_KEY)

        http = Http(
            url=ax_url,
            cert_client_cert=cert_path,
            cert_client_key=key_path,
            certwarn=False,
        )

        http()

    def test_client_cert_both(self, request, tmp_path):
        """Test cert or key supplied, but not the other."""
        ax_url = get_url(request)
        both_path = tmp_path / TEST_CLIENT_CERT_NAME
        both_cert = "{}\n{}\n".format(TEST_CLIENT_CERT, TEST_CLIENT_KEY)
        both_path.write_text(both_cert)

        http = Http(url=ax_url, cert_client_both=both_path, certwarn=False)

        http()

    def test_log_req_attrs_true(self, request, caplog):
        """Test verbose logging of request attrs when log_request_attrs=True."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(
            url=ax_url,
            log_request_attrs=["url", "headers"],
            certwarn=False,
            log_level="debug",
        )

        http()

        assert len(caplog.records) == 1

        entries = ["REQUEST ATTRS:.*{}.*headers".format(http.url)]
        log_check(caplog, entries)

    def test_log_req_attrs_none(self, request, caplog):
        """Test no logging of request attrs when log_request_attrs=None."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(
            url=ax_url, log_request_attrs=None, certwarn=False, log_level="debug"
        )

        http()

        assert not caplog.records

    def test_log_resp_attrs_true(self, request, caplog):
        """Test verbose logging of response attrs when log_response_attrs=True."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(
            url=ax_url,
            log_response_attrs=["url", "headers"],
            certwarn=False,
            log_level="debug",
        )

        http()

        assert len(caplog.records) == 1

        entries = ["RESPONSE ATTRS:.*{}.*headers".format(http.url)]
        log_check(caplog, entries)

    def test_log_response_attrs_all(self, request, caplog):
        """Test no logging of response attrs when log_response_attrs=None."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(
            url=ax_url, log_response_attrs="all", certwarn=False, log_level="debug"
        )

        http()

        assert caplog.records

    def test_log_response_attrs_none(self, request, caplog):
        """Test no logging of response attrs when log_response_attrs=None."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(
            url=ax_url, log_response_attrs=None, certwarn=False, log_level="debug"
        )

        http()

        assert not caplog.records

    def test_log_resp_body_true(self, request, caplog):
        """Test logging of response body when log_response_body=True."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(
            url=ax_url, log_response_body=True, certwarn=False, log_level="debug"
        )

        http()

        assert len(caplog.records) == 1

        entries = ["RESPONSE BODY:.*"]
        log_check(caplog, entries)

    def test_log_resp_body_false(self, request, caplog):
        """Test no logging of response body when log_response_body=False."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(
            url=ax_url, log_response_body=False, certwarn=False, log_level="debug"
        )

        http()

        assert not caplog.records

    def test_log_req_body_true(self, request, caplog):
        """Test logging of request body when log_request_body=True."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(url=ax_url, log_request_body=True, certwarn=False, log_level="debug")

        http()

        assert len(caplog.records) == 1

        entries = ["REQUEST BODY:.*"]
        log_check(caplog, entries)

    def test_log_req_body_false(self, request, caplog):
        """Test no logging of request body when log_request_body=False."""
        caplog.set_level(logging.DEBUG)

        ax_url = get_url(request)

        http = Http(
            url=ax_url, log_request_body=False, certwarn=False, log_level="debug"
        )

        http()

        assert not caplog.records



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
import logging
import pathlib
import time

import pytest

from axonius_api_client.constants import (
    LOG_LEVEL_CONSOLE,
    LOG_LEVEL_FILE,
    LOG_NAME_FILE,
    LOG_NAME_STDERR,
    LOG_NAME_STDOUT,
)
from axonius_api_client.exceptions import ToolsError
from axonius_api_client.logs import (
    LOG,
    add_file,
    add_null,
    add_stderr,
    add_stdout,
    del_file,
    del_null,
    del_stderr,
    del_stdout,
    get_obj_log,
    gmtime,
    localtime,
    str_level,
)


class TestLogs:
    """Test logs."""

    def test_gmtime(self):
        """Pass."""
        gmtime()
        assert logging.Formatter.converter == time.gmtime

    def test_localtime(self):
        """Pass."""
        localtime()
        assert logging.Formatter.converter == time.localtime

    def test_get_obj_log(self):
        """Pass."""
        log = get_obj_log(obj=self, level="warning")
        assert log.name == "axonius_api_client.tests.tests_pkg.test_logs.TestLogs"
        assert log.level == logging.WARNING

    def test_str_level_int(self):
        """Pass."""
        assert str_level(level=10) == "DEBUG"

    def test_str_level_str_int(self):
        """Pass."""
        assert str_level(level="10") == "DEBUG"

    def test_str_level_str(self):
        """Pass."""
        assert str_level(level="debug") == "DEBUG"

    def test_str_level_fail(self):
        """Pass."""
        with pytest.raises(ToolsError):
            str_level(level="xx")

    def test_add_del_stderr(self):
        """Pass."""
        h = add_stderr(obj=LOG)
        assert h.name == LOG_NAME_STDERR
        assert str_level(level=h.level).lower() == LOG_LEVEL_CONSOLE
        assert isinstance(h, logging.StreamHandler)
        assert h in LOG.handlers

        dh = del_stderr(obj=LOG)
        assert isinstance(dh, dict)
        assert LOG.name in dh
        assert isinstance(dh[LOG.name], list)
        assert h in dh[LOG.name]
        assert h not in LOG.handlers

    def test_add_del_stdout(self):
        """Pass."""
        h = add_stdout(obj=LOG)
        assert h.name == LOG_NAME_STDOUT
        assert str_level(level=h.level).lower() == LOG_LEVEL_CONSOLE
        assert isinstance(h, logging.StreamHandler)
        assert h in LOG.handlers

        dh = del_stdout(obj=LOG)
        assert isinstance(dh, dict)
        assert LOG.name in dh
        assert isinstance(dh[LOG.name], list)
        assert h in dh[LOG.name]
        assert h not in LOG.handlers

    def test_add_del_null(self):
        """Pass."""
        del_null(obj=LOG)
        h = add_null(obj=LOG)
        assert h.name == "NULL"
        assert isinstance(h, logging.NullHandler)
        assert h in LOG.handlers

        fh = add_null(obj=LOG)
        assert fh is None

        dh = del_null(obj=LOG)

        assert isinstance(dh, dict)
        assert isinstance(dh[LOG.name], list)

        assert LOG.name in dh
        f = dh.pop(LOG.name)

        assert h in f
        assert h not in LOG.handlers
        assert not dh

    def test_add_del_file(self):
        """Pass."""
        h = add_file(obj=LOG)
        assert h.name == LOG_NAME_FILE
        assert str_level(level=h.level).lower() == LOG_LEVEL_FILE
        assert isinstance(h, logging.handlers.RotatingFileHandler)
        assert h in LOG.handlers
        assert getattr(h, "PATH", None)
        assert isinstance(h.PATH, pathlib.Path)

        dh = del_file(LOG)
        assert isinstance(dh, dict)
        assert LOG.name in dh
        assert isinstance(dh[LOG.name], list)
        assert h in dh[LOG.name]
        assert h not in LOG.handlers



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, NODE_CNX, add_options
from .grp_common import EXPORT, handle_export

OPTIONS = [
    *AUTH,
    EXPORT,
    *NODE_CNX,
]


@click.command(name="get", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, **kwargs):
    """Get connections."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = client.adapters.cnx.get_by_adapter(**kwargs)
    handle_export(ctx=ctx, rows=rows, export_format=export_format)



# -*- coding: utf-8 -*-
"""Test suite for axonapi.api.enforcements."""
import pytest

from ...meta import (DEPLOY_ACTION_NAME, DEPLOY_FILE_CONTENTS,
                     DEPLOY_FILE_NAME, LINUX_QUERY, SHELL_ACTION_CMD,
                     SHELL_ACTION_NAME)


class TestRunActionBase:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_run_action):
        """Pass."""
        return api_run_action


class TestRunActionPrivate(TestRunActionBase):
    """Pass."""

    def test_private_get(self, apiobj):
        """Pass."""
        data = apiobj._get()
        for i in ["deploy", "shell", "upload_file"]:
            assert i in data

    # this returns nothing...
    # AND no action shows up in GUI for dvc
    # AND no task shows up in EC
    # BUT: Extended Data Tab shows stuff, but i dont know how to query for that yet
    def test_private_shell(self, apiobj, api_devices):
        """Pass."""
        devices = api_devices._get(query=LINUX_QUERY, page_size=1, row_start=0)
        ids = [x["internal_axon_id"] for x in devices["assets"]]

        if not ids:
            pytest.skip("No linux devices found!")

        data = apiobj._shell(
            action_name=SHELL_ACTION_NAME, ids=ids, command=SHELL_ACTION_CMD,
        )
        assert not data

    @pytest.fixture(scope="class")
    def uploaded_file(self, apiobj):
        """Pass."""
        data = apiobj._upload_file(name=DEPLOY_FILE_NAME, content=DEPLOY_FILE_CONTENTS)
        assert isinstance(data, dict)
        assert isinstance(data["uuid"], str)
        assert data["filename"] == DEPLOY_FILE_NAME
        return data

    # returns nadda
    def test_private_upload_deploy(self, apiobj, api_devices, uploaded_file):
        """Pass."""
        devices = api_devices._get(query=LINUX_QUERY, page_size=1, row_start=0)
        ids = [x["internal_axon_id"] for x in devices["assets"]]

        if not ids:
            reason = "No linux devices found!"
            pytest.skip(reason)

        data = apiobj._deploy(
            action_name=DEPLOY_ACTION_NAME,
            ids=ids,
            file_uuid=uploaded_file["uuid"],
            file_name=uploaded_file["filename"],
            params=None,
        )
        assert not data



# -*- coding: utf-8 -*-
"""Test suite for assets."""
import copy
import io

import pytest
from axonius_api_client.constants import AGG_ADAPTER_NAME

from .callbacks import Callbacks, load_test_data


class CallbacksCsv(Callbacks):
    """Pass."""

    @pytest.fixture(scope="class")
    def cbexport(self):
        """Pass."""
        return "csv"

    def test_row_as_is(self, cbexport, apiobj):
        """Pass."""
        field_complex = apiobj.TEST_DATA["field_complex"]
        if field_complex is None:
            pytest.skip(f"No complex field found for {apiobj}")

        schema = apiobj.fields.get_field_schema(
            value=field_complex,
            schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME],
        )
        sub_titles = [x["column_title"] for x in schema["sub_fields"] if x["is_root"]]

        io_fd = io.StringIO()

        getargs = {"export_fd": io_fd}

        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.start()

        start_val = io_fd.getvalue().splitlines()[0]
        for i in sub_titles:
            assert f'"{i}"' in start_val

        for row in copy.deepcopy(apiobj.TEST_DATA["cb_assets"][:200]):
            row_id = row["internal_axon_id"]
            rows_ret = cbobj.row(row=copy.deepcopy(row))
            assert len(rows_ret) == 1
            assert rows_ret[0] == {"internal_axon_id": row_id}

        cbobj.stop()
        output = io_fd.getvalue()
        assert output.endswith("\n\n")


class TestDevicesCallbacksCsv(CallbacksCsv):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_devices):
        """Pass."""
        return load_test_data(apiobj=api_devices)


class TestUsersCallbacksCsv(CallbacksCsv):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_users):
        """Pass."""
        return load_test_data(apiobj=api_users)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...context import AliasedGroup
from . import (cmd_add, cmd_add_from_json, cmd_delete_by_name,
               cmd_delete_by_tags, cmd_get, cmd_get_by_name, cmd_get_by_tags,
               cmd_get_tags)


@click.group(cls=AliasedGroup)
def saved_query():
    """Group: Work with saved queries."""


saved_query.add_command(cmd_add.cmd)
saved_query.add_command(cmd_add_from_json.cmd)
saved_query.add_command(cmd_delete_by_name.cmd)
saved_query.add_command(cmd_delete_by_tags.cmd)
saved_query.add_command(cmd_get.cmd)
saved_query.add_command(cmd_get_tags.cmd)
saved_query.add_command(cmd_get_by_name.cmd)
saved_query.add_command(cmd_get_by_tags.cmd)
